__all__ = ('TestRunner', 'run_tests_in')

import sys
from functools import partial as partial_func
from os import getcwd as get_current_working_directory
from os.path import sep as PATH_SEPARATOR

from scarletio import RichAttributeErrorBaseType, include, render_exception_into

from ... import __package__ as PACKAGE_NAME

from ..environment import EnvironmentManager
from ..events import (
    FileLoadDoneEvent, FileRegistrationEvent, FileRegistrationDoneEvent, FileTestingDoneEvent, SourceLoadFailureEvent,
    TestDoneEvent, TestingEndEvent, TestingStartEvent
)
from ..file import FileSystemEntry, iter_collect_test_files_in

from .context import RunnerContext


create_default_event_handler_manager = include('create_default_event_handler_manager')


def setup_test_library_import():
    """
    Setups test directory import if not on path instead running it relatively.
    """
    split = PACKAGE_NAME.split('.')
    if len(split) <= 1:
        return None
    
    module = __import__(PACKAGE_NAME)
    
    for directory_name in split[1:]:
        module = module.__dict__[directory_name]
    
    sys.modules[split[-1]] = module


def _add_to_system_path_callback(path, runner):
    """
    Removes the given path from `sys.path`.
    
    Parameters
    ----------
    path : `str`
        The path to remove.
    runner : ``TestRunner``
        The respective test runner.
    """
    if path not in sys.path:
        sys.path.append(path)
        
    

def _remove_from_system_path_callback(path, runner):
    """
    Removes the given path from `sys.path`.
    
    Parameters
    ----------
    path : `str`
        The path to remove.
    runner : ``TestRunner``
        The respective test runner.
    """
    try:
        sys.path.remove(path)
    except ValueError:
        pass


def _ignore_test_call_frame(frame):
    """
    Ignores test runner frames when rendering event handler exception
    
    Parameters
    ----------
    frame : ``FrameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.file_name
    name = frame.name
    line = frame.line
    
    if file_name == __file__:
        if name == 'run':
            if line == 'event_handler(event)':
                should_show_frame = False
    
    return should_show_frame


def _render_event_exception(event_handler, event, err):
    """
    Renders exception occurred inside of en event handler.
    
    Parameters
    ----------
    event_handler : `callable`
        The event handler which failed to run.
    
    event : ``EventBase``
        The dispatched event.
    
    err : `BaseException`
        The occurred exception.
    """
    
    exception_message_parts = [
        'Exception occurred meanwhile running event handler: ',
        repr(event_handler),
        ' with event ',
        repr(event),
        '\n\n'
    ]
    
    render_exception_into(err, exception_message_parts, filter = _ignore_test_call_frame)
    
    return ''.join(exception_message_parts)


class TestRunner(RichAttributeErrorBaseType):
    """
    Attributes
    ----------
    _path_parts : `None | list<str>`
        Added path parts to specify from which which directory we want to collect the tests from.
    _source_directory : `str`
        The path to run tests from.
    sources : `set<str>`
        Sources to import before executing any test.
    _stopped : `bool`
        Whether testing is stopped.
    _teardown_callbacks : `None`, `list` of `callable`
        Functions to call when the tests are finished.
    environment_manager : ``EnvironmentManager``
        Testing environment manager.
    event_handler_manager : ``EventHandlerManager``
        Event handler container.
    
    Utility Methods
    ---------------
    - ``.run``
    - ``.stop``
    - ``.add_teardown_callback``
    """
    __slots__ = (
        '_path_parts', '_source_directory', '_sources', '_stopped', '_teardown_callbacks', 'environment_manager',
        'event_handler_manager'
    )
    
    def __new__(
        cls, source_directory, sources, path_parts = None, *, environment_manager = None, event_handler_manager = None
    ):
        """
        Parameters
        ----------
        source_directory : `str`
            The path to run tests from.
        sources : `set<str>`
            Sources to import before executing any test.
        path_parts : `None | list<str>` = `None`, Optional
            Added path parts to specify from which which directory we want to collect the tests from.
        environment_manager : `None`, ``EnvironmentManager`` = `None`, Optional (Keyword only)
            Testing environment manager.
        event_handler_manager : `None`, ``EventHandlerManager`` = `None`, Optional (Keyword only)
            Event handler container.
        """
        if (path_parts is not None):
            if path_parts:
                path_parts = path_parts.copy()
            else:
                path_parts = None
            
        if environment_manager is None:
            environment_manager = EnvironmentManager()
        
        environment_manager = environment_manager.populate()
        
        if event_handler_manager is None:
            event_handler_manager = create_default_event_handler_manager()
        
        self = object.__new__(cls)
        self._path_parts = path_parts
        self._source_directory = source_directory
        self._sources = sources
        self._stopped = False
        self._teardown_callbacks = None
        self.environment_manager = environment_manager
        self.event_handler_manager = event_handler_manager
        return self
    
    
    def _setup(self):
        """
        Setups the test dependencies.
        """
        # Add source directory to sys.path if not there
        source_directory = self._source_directory
        
        if source_directory in sys.path:
            source_directory_in_system_paths = True
        else:
            sys.path.append(source_directory)
            source_directory_in_system_paths = False
        
        if not source_directory_in_system_paths:
            self.add_teardown_callback(partial_func(_remove_from_system_path_callback, source_directory))
        
        # Remove working directory from sys.path if under source directory
        working_directory = get_current_working_directory()
        if (
            working_directory in sys.path and
            len(working_directory) > len(source_directory) and
            working_directory.startswith(source_directory) and
            working_directory[len(source_directory)] == PATH_SEPARATOR
        ):
            working_directory_under_source = True
            try:
                sys.path.remove(working_directory)
            except ValueError:
                pass
        else:
            working_directory_under_source = False
        
        if not working_directory_under_source:
            self.add_teardown_callback(partial_func(_add_to_system_path_callback, working_directory))
        
        setup_test_library_import()
    
    
    def add_teardown_callback(self, callback):
        """
        Adds a callback to run when the tests are finished.
        
        Parameters
        ----------
        callback : `callable`
            The callback to add.
        """
        teardown_callbacks = self._teardown_callbacks
        if (teardown_callbacks is None):
            teardown_callbacks = []
            self._teardown_callbacks = teardown_callbacks
        
        teardown_callbacks.append(callback)
    
    
    def _exhaust_teardown_callbacks(self):
        """
        Runs the teardown callbacks.
        """
        teardown_callbacks = self._teardown_callbacks
        self._teardown_callbacks = None
        
        if (teardown_callbacks is not None):
            while teardown_callbacks:
                teardown_callbacks.pop()(self)
    
    
    def _teardown(self):
        """
        Clears up after the tests ran.
        """
        self._exhaust_teardown_callbacks()
    
    
    def _run_generator(self):
        """
        Runs the tests of the runner. Witch each iteration step it drops back the created event.
        
        This method is an iterable generator.
        
        Yields
        ------
        event : ``EventBase``
        """
        try:
            # Setup
            self._setup()
            
            # Build context.
            path_parts = self._path_parts
            if path_parts is None:
                file_system_entries = [
                    FileSystemEntry(self._source_directory, source, None) for source in self._sources
                ]
            else:
                file_system_entries = [FileSystemEntry(self._source_directory, path_parts[0], path_parts[1:])]
            
            context = RunnerContext(self, file_system_entries)
            
            # Import plugins.
            for source in self._sources:
                try:
                    __import__(source)
                except BaseException as err:
                    yield SourceLoadFailureEvent(context, source, err)
                    return
            
            yield TestingStartEvent(context)
            
            # Collect test files
            for file_system_entry in context.file_system_entries:
                for test_file in iter_collect_test_files_in(file_system_entry):
                    context.register_file(test_file)
                    yield FileRegistrationEvent(context, test_file)
            
            yield FileRegistrationDoneEvent(context)
            
            # Load test files
            for registered_file in context.iter_registered_files_shallow():
                for test_file in registered_file.iter_test_files():
                    if test_file.is_directory():
                        test_file.get_module()
                        
                        yield FileLoadDoneEvent(context, test_file)
                        
                        if test_file.is_loaded_with_failure():
                            break
                        
                        yield FileTestingDoneEvent(context, test_file)
                    
                    else:
                        test_file.try_load_test_cases()
                        
                        yield FileLoadDoneEvent(context, test_file)
                        
                        # Run test file if loaded successfully
                        if test_file.is_loaded_with_success():
                            
                            for result in test_file.iter_invoke_test_cases(self.environment_manager):
                                yield TestDoneEvent(context, result)
                            
                            yield FileTestingDoneEvent(context, test_file)
            
            yield TestingEndEvent(context)
        
        finally:
            self._teardown()
    
    
    def run(self):
        """
        Runs the tests of the test runner.
        """
        if self._stopped:
            return
        
        event_handler_manager = self.event_handler_manager
        
        for event in self._run_generator():
            for event_handler in event_handler_manager.iter_handlers_for_event(event):
                try:
                    event_handler(event)
                except (KeyboardInterrupt, SystemExit):
                    sys.stderr.write('Shutting down test runner...\n')
                    sys.stderr.flush()
                    raise
                
                except BaseException as err:
                    sys.stderr.write(_render_event_exception(event_handler, event, err))
            
            if self._stopped:
                break
    
    
    def stop(self):
        """
        Stops the testing step loop.
        """
        self._stopped = True


def run_tests_in(source_directory, sources, test_collection_route):
    """
    Runs tests from the given `source_directory` and collects them from the given `test_collection_route`.
    Or from `sources` if not specified.
    
    Parameters
    ----------
    source_directory : `str`
        The path to import the sources from.
    sources : `set<str>`
        Sources to import before executing any test.
    test_collection_route : `None | list<str>`
        Added path parts to specify from which which directory we want to collect the tests from.
    """
    TestRunner(source_directory, sources, test_collection_route).run()



__all__ = ()

def maybe_merge_iterables(iterable_1, iterable_2):
    """
    Merges the two maybe iterable if applicable returning a new one.
    
    Parameters
    ----------
    iterable_1 : `None`, `iterable`
        Iterable to merge.
    iterable_2 : `None`, `iterable`
        Iterable to merge.
    
    Returns
    -------
    merged : `None`, `list`
    """
    if (iterable_1 is not None) and (not iterable_1):
        iterable_1 = None
    
    if (iterable_2 is not None) and (not iterable_2):
        iterable_2 = None
    
    if iterable_1 is None:
        if iterable_2 is None:
            merged = None
        else:
            merged = [*iterable_2]
    else:
        if iterable_2 is None:
            merged = [*iterable_1]
        else:
            merged = [*iterable_1, *iterable_2]
    
    return merged


def maybe_merge_mappings(mapping_1, mapping_2):
    """
    Merges the two maybe mapping if applicable returning a new one.
    
    Parameters
    ----------
    mapping_1 : `None`, `mapping`
        Mapping to merge.
    mapping_2 : `None`, `mapping`
        Mapping to merge.
    
    Returns
    -------
    merged : `None`, `dict`
    """
    if (mapping_1 is not None) and (not mapping_1):
        mapping_1 = None
    
    if (mapping_2 is not None) and (not mapping_2):
        mapping_2 = None
    
    if mapping_1 is None:
        if mapping_2 is None:
            merged = None
        else:
            merged = {**mapping_2}
    else:
        if mapping_2 is None:
            merged = {**mapping_1}
        else:
            merged = {**mapping_1, **mapping_2}
    
    return merged



from .base import *
from .constants import *
from .file_load_done import *
from .file_registration import *
from .file_registration_done import *
from .file_testing_done import *
from .source_load_failure import *
from .test_done import *
from .testing_end import *
from .testing_start import *


__all__ = (
    *base.__all__,
    *constants.__all__,
    *file_load_done.__all__,
    *file_registration.__all__,
    *file_registration_done.__all__,
    *file_testing_done.__all__,
    *source_load_failure.__all__,
    *test_done.__all__,
    *testing_end.__all__,
    *testing_start.__all__,
)



__all__ = ('AssertionInstance',)

from scarletio import copy_docs

from ..helpers.un_nesting import un_nest_types

from .assertion_conditional_base import (
    AssertionConditionalBase, AssertionConditionalBase2Value, _render_parameters_representation_into,
    _render_types_parameter_representation_into
)


class AssertionInstance(AssertionConditionalBase2Value):
    """
    Asserts whether the first object is instance of the second one.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        First value to assert instance with.
    value_1 : `set` of `type<type>`
        The second value to assert instance with.
    accept_subtypes : `bool`
        Whether instances of subtypes should be accepted.
    nullable : `bool`
        Whether `value` is accepted even if given as `None`.
    """
    __slots__ = ('accept_subtypes', 'nullable')
    
    def __new__(cls, value, accepted_type, *accepted_types, accept_subtypes = True, reverse = False, nullable = False):
        """
        Creates a new instance assertion.
        
        Parameters
        ----------
        value : `object`
            Object to check.
        accepted_type : `type<type>`
            The type to check.
        *accepted_types : `tuple<type<type>, ...>`
            Additional accepted types.
        accept_subtypes : `bool` = `True`, Optional (Keyword only)
            Whether instances of subtypes should be accepted.
        reverse : `bool` = `False`, Optional (Keyword only)
            Whether the condition should be reversed.
        nullable : `bool` = `False`, Optional (Keyword only)
            Whether `value` is accepted even if given as `None`.
        
        Raises
        ------
        TypeError
            - If a type`` is not `type` instance.
        ValueError
            - No types given.
        """
        accepted_types = un_nest_types((accepted_type, *accepted_types))
        if not accepted_types:
            raise ValueError(
                'At least 1 type is required.'
            )
        
        self = AssertionConditionalBase2Value.__new__(cls, value, accepted_types, reverse = reverse)
        self.accept_subtypes = accept_subtypes
        self.nullable = nullable
        return self
    
    
    @copy_docs(AssertionConditionalBase2Value.invoke_condition)
    def invoke_condition(self):
        value = self.value_0
        accepted_types = self.value_1
        
        if self.nullable:
            if value is None:
                return True
        
        value_type = type(value)
        
        if self.accept_subtypes:
            return any(issubclass(value_type, accepted_type) for accepted_type in accepted_types)
        
        return any((value_type is accepted_type) for accepted_type in accepted_types)
    
    
    @copy_docs(AssertionConditionalBase2Value._get_operation_representation)
    def _get_operation_representation(self):
        return 'instance'
    
    
    @copy_docs(AssertionConditionalBase2Value._render_operation_representation_into)
    def _render_operation_representation_into(self, into):
        AssertionConditionalBase._render_operation_representation_into(self, into)
        
        into.append(' as "')
        
        if self.nullable:
            into.append('value is None or ')
        
        if self.accept_subtypes:
            instance_check_part = 'isinstance(value, expected_types)'
        else:
            instance_check_part = 'type(value) is expected_types'
        into.append(instance_check_part)
        into.append('"')
        
        return into
    
    
    @copy_docs(AssertionConditionalBase2Value._build_repr_parts_into)
    def _build_repr_parts_into(self, into):
        into = AssertionConditionalBase2Value._build_repr_parts_into(self, into)
        
        accept_subtypes = self.accept_subtypes
        if not accept_subtypes:
            into.append(', accept_subtypes = ')
            into.append(repr(accept_subtypes))
        
        nullable = self.nullable
        if nullable:
            into.append(', nullable = ')
            into.append(repr(nullable))
        
        return into
    
    
    @copy_docs(AssertionConditionalBase2Value.render_failure_message_parts_into)
    def render_failure_message_parts_into(self, failure_message_parts):
        AssertionConditionalBase.render_failure_message_parts_into(self, failure_message_parts)
        _render_parameters_representation_into('value', self.value_0, failure_message_parts)
        _render_types_parameter_representation_into('expected_types', self.value_1, failure_message_parts)
        return failure_message_parts



__all__ = ()

from os.path import sep as PATH_SEPARATOR


PATH_REPR_MAX_PARTS = 4
PATH_REPR_PART_PLACEHOLDER = '.'
PATH_REPR_MAX_LENGTH = 80


def get_short_path_repr(path):
    """
    Returns the short representation of the given path.
    
    Parameters
    ----------
    path : `str`
        The path to get it's short representation of.
    
    Returns
    -------
    path_repr : `str`
        The path's representation.
    """
    path_parts = path.split(PATH_SEPARATOR)
    
    path_parts_to_render = []
    
    collective_length = 0
    added_field_count = 0
    was_last_field_last = False
    was_last_followed_by_empty = False
    
    
    while path_parts:
        path_part = path_parts.pop()
        if (collective_length < PATH_REPR_MAX_LENGTH) and (added_field_count < PATH_REPR_MAX_PARTS):
            new_length = collective_length + len(path_part)
            if new_length > PATH_REPR_MAX_LENGTH:
                if added_field_count == 0:
                    path_parts_to_render.append(path_part)
                else:
                    path_parts_to_render.append(PATH_SEPARATOR)
                    path_parts_to_render.append('.')
                was_last_field_last = True
            
            else:
                if added_field_count:
                    path_parts_to_render.append(PATH_SEPARATOR)
                
                added_field_count += 1
                if (added_field_count == PATH_REPR_MAX_PARTS):
                    was_last_field_last = True
                
                path_parts_to_render.append(path_part)
            
            collective_length = new_length
        
        else:
            if was_last_field_last:
                was_last_field_last = False
                
                path_parts_to_render.append(PATH_SEPARATOR)
                
                if path_part:
                    path_parts_to_render.append('.')
                else:
                    was_last_followed_by_empty = True
            else:
                if was_last_followed_by_empty:
                    was_last_followed_by_empty = False
                    path_parts_to_render.append(PATH_SEPARATOR)
                    
                if path_part:
                    path_parts_to_render.append('.')
    
    
    path_parts_to_render.reverse()
    return ''.join(path_parts_to_render)



__all__ = ('EventHandlerManager',)

from scarletio import CallableAnalyzer, RichAttributeErrorBaseType

from ..events import EventBase


class EventHandlerManager(RichAttributeErrorBaseType):
    """
    Event handler for handling test runner events.
    
    Attributes
    -----------
    _event_handlers_by_identifier : `dict` of (`int`, `callable`) items
        The registered event handlers.
    """
    __slots__ = ('_event_handlers_by_identifier',)
    
    def __new__(cls):
        """
        Creates a new event handler.
        """
        self = object.__new__(cls)
        self._event_handlers_by_identifier = {}
        return self
    
    
    def events(self, event_handler):
        """
        Can be sued to register event handlers by either using it as a decorator, or just calling it.
        
        Parameters
        ----------
        event_handler : `callable`
            the function to add as an event handler.
        
        Returns
        -------
        event_handler : `callable`
        
        Raises
        ------
        TypeError
            - If `event`'s types is incorrect.
        """
        analyzer = CallableAnalyzer(event_handler)
        
        event_parameter = None
        
        for parameter in analyzer.iter_non_reserved_parameters():
            if parameter.is_keyword_only() or parameter.is_args() or parameter.is_kwargs():
                raise TypeError(
                    f'Event handlers do not support keyword only, args or kwargs parameters, '
                    f'got event_handler={event_handler!r} with parameter={parameter!r}.'
                )
            
            if event_parameter is None:
                event_parameter = parameter
            
            else:
                if not parameter.has_default():
                    raise TypeError(
                        f'To event handlers only 1 parameter is passed, but the given one expects more, '
                        f'got event_handler={event_handler!r} with parameter={parameter!r}.'
                    )
        
        
        if (event_parameter is None):
            raise TypeError(
                f'To event handlers 1 parameter is passed, but the given one expects zero, '
                f'got event_handler={event_handler!r}.'
            )
        
        if not event_parameter.has_annotation:
            raise TypeError(
                'Event handler\'s event parameter should be annotated with an event\'s type, '
                f'got event_handler={event_handler!r} with parameter={event_parameter!r}.'
            )
        
        annotation = event_parameter.annotation
        if (not isinstance(annotation, type)) or (not issubclass(annotation, EventBase)):
            raise TypeError(
                'Event handler\'s event parameter should be annotated with an event\'s type, '
                f'got event_handler={event_handler!r} with parameter={event_parameter!r}; annotation={annotation!r}.'
            )
        
        if annotation.identifier == EventBase.identifier:
            raise TypeError(
                'Event handler\'s event parameter is annotated with a base event type, '
                f'got event_handler={event_handler!r} with parameter={event_parameter!r}; annotation={annotation!r}.'
            )
        
        # Register event handler
        event_handlers_by_identifier = self._event_handlers_by_identifier
        try:
            event_handlers = event_handlers_by_identifier[annotation.identifier]
        except KeyError:
            event_handlers = []
            event_handlers_by_identifier[annotation.identifier] = event_handlers
        
        event_handlers.append(event_handler)
        
        
        return event_handler
    
    
    def iter_handlers_for_event(self, event):
        """
        Iterates over the event handler which should be called with the given event.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        event : ``EventBase``
            Event instance.
        
        Yields
        ------
        event_handler : `callable`
        """
        try:
            event_handlers = self._event_handlers_by_identifier[event.identifier]
        except KeyError:
            pass
        else:
            yield from event_handlers
    
    
    def copy(self):
        """
        Copies the event handler manager returning a new one.
        
        Returns
        -------
        new : ``EventHandlerManager``
        """
        event_handlers_by_identifier = {}
        
        for event_identifier, event_handlers in self._event_handlers_by_identifier.items():
            event_handlers_by_identifier[event_identifier] = event_handlers.copy()
        
        new = object.__new__(type(self))
        new._event_handlers_by_identifier = event_handlers_by_identifier
        return new



from .asserting import *
from .output import *
from .base import *
from .parameter_mismatch import *
from .raising import *
from .returning import *


__all__ = (
    *asserting.__all__,
    *base.__all__,
    *parameter_mismatch.__all__,
    *output.__all__,
    *raising.__all__,
    *returning.__all__,
)



from .exception_matching import *
from .hashing import *
from .merging import *
from .path_repr import *
from .un_nesting import *


__all__ = (
    *exception_matching.__all__,
    *hashing.__all__,
    *merging.__all__,
    *path_repr.__all__,
    *un_nesting.__all__,
)



__all__ = ('AssertionConditionalBase', 'AssertionConditionalBase1Value', 'AssertionConditionalBase2Value')

from scarletio import copy_docs, include

from .assertion_base import AssertionBase
from .assertion_states import ASSERTION_STATE_CREATED, ASSERTION_STATE_PASSED, ASSERTION_STATE_FAILED


AssertionException = include('AssertionException')


class InvokeAfterConstructor(type):
    """
    Metatype for invoking a test after calling its `__new__`.
    """
    def __call__(type_, *positional_parameters, **keyword_parameter):
        """
        Instantiates the type and invokes it.
        
        Parameters
        ----------
        type_ : `type`
            The type to Instantiate and invoke.
        *positional_parameters : `tuple<object>`
            Positional parameters to pass to the type's constructor.
        keyword_parameter : `dict<str, object>`
            Keyword parameters to pass to the type's constructor.
        
        Returns
        -------
        condition_return : `object`
            The returned value by the condition. Usually a boolean.
        
        Raises
        ------
        AssertionException
            The condition failed.
        BaseException
            An other occurred exceptions.
        """
        return type_.__new__(type_, *positional_parameters, **keyword_parameter).invoke()


class AssertionConditionalBase(AssertionBase, metaclass = InvokeAfterConstructor):
    """
    Base class for conditional assertions.
    
    Attributes
    ----------
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    state : `str`
        The condition's state.
    """
    __slots__ = ('exception', 'reverse')
    
    def __new__(cls, *, reverse = False):
        """
        Creates an new conditional assertion instance.
        
        Parameters
        ----------
        reverse : `bool` = `False`, Optional (Keyword only)
            Whether the condition should be reversed.
        """
        self = AssertionBase.__new__(cls)
        self.exception = None
        self.reverse = reverse
        self.state = ASSERTION_STATE_CREATED
        return self
    
    
    @copy_docs(AssertionBase._build_repr_parts_into)
    def _build_repr_parts_into(self, into):
        into = AssertionBase._build_repr_parts_into(self, into)
        
        reverse = self.reverse
        if reverse:
            into.append(', reverse = ')
            into.append(repr(reverse))
            
        
        exception = self.exception
        if (exception is not None):
            into.append(', exception = ')
            into.append(repr(exception))
        
        return into
    
    
    def invoke(self):
        """
        Invokes the assertion.
        
        Returns
        -------
        condition_return : `object`
            The value returned by the condition.
        
        Raises
        ------
        AssertionException
            The condition failed.
        """
        try:
            condition_return = self.invoke_condition()
            
            if condition_return:
                passed = True
            else:
                passed = False
        
        except BaseException as err:
            self.state = ASSERTION_STATE_FAILED
            self.exception = err
            
        else:
            if self.reverse:
                passed = not passed
            
            if passed:
                self.state = ASSERTION_STATE_PASSED
                return condition_return
            
            self.state = ASSERTION_STATE_FAILED
        
        try:
            raise AssertionException(self)
        finally:
            # Remove self reference, so garbage collection wont fail
            self = None
    
    
    def invoke_condition(self):
        """
        Invokes the condition.
        
        Returns
        -------
        result : `object`
            The condition's result.
        """
        raise NotImplementedError
    
    
    def _get_operation_representation(self):
        """
        Gets the assertion's operation's representation.
        
        Returns
        -------
        operation_representation : `str`
        """
        return ''
    
    
    def _render_operation_representation_into(self, into):
        """
        Adds the operation representation to the given list.
        
        Parameters
        ----------
        into : `list` of `str`
            A list to extend with the rendered strings.
        
        Returns
        -------
        into : `list` of `str`
        """
        into.append('Operation: "')
        into.append(self._get_operation_representation())
        into.append('"')
        
        if self.reverse:
            into.append(' (reversed)')
        
        return into
    
    
    @copy_docs(AssertionBase.render_failure_message_parts_into)
    def render_failure_message_parts_into(self, failure_message_parts):
        self._render_operation_representation_into(failure_message_parts)
        failure_message_parts.append('\n')
        return failure_message_parts


def _render_parameters_representation_into(parameter_name, parameter_value, into):
    """
    Renders the given parameter into the given list of strings.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter's name.
    parameter_value : `object`
        The parameter's value.
    into : `list` of `str`
        A list to extend with the rendered strings.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(parameter_name)
    into.append(' = ')
    into.append(repr(parameter_value))
    into.append('\n')
    
    return into


def _render_types_parameter_representation_into(parameter_name, types, into):
    """
    Renders the given types parameter into the given list of strings.
    
    Parameters
    ----------
    parameter_name : `str`
        The parameter's name.
    types : `set` of `type`
        The parameter's value.
    into : `list` of `str`
        A list to extend with the rendered strings.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(parameter_name)
    into.append(' = ')
    
    type_representations = sorted(str(type.__name__) for type in types)
    length = len(type_representations)
    if length:
        index = 0
        
        while True:
            type_representation = type_representations[index]
            into.append(type_representation)
            
            index += 1
            if index == length:
                break
            
            into.append(', ')
            continue
    
    into.append('\n')
    
    return into


class AssertionConditionalBase1Value(AssertionConditionalBase):
    """
    Base class for executing a one value assertion.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        The value to call the condition on.
    """
    __slots__ = ('value_0',)

    def __new__(cls, value_0, *, reverse = False):
        """
        Asserts whether the two values are equal. Fails the test if not.
        
        Parameters
        ----------
        value_0 : `object`
            First value to assert equality with.
        reverse : `bool` = `False`, Optional (Keyword only)
            Whether the condition should be reversed.
        """
        self = AssertionConditionalBase.__new__(cls, reverse = reverse)
        self.value_0 = value_0
        return self
    
    
    @copy_docs(AssertionConditionalBase._build_repr_parts_into)
    def _build_repr_parts_into(self, into):
        into = AssertionConditionalBase._build_repr_parts_into(self, into)
        
        into.append(', value_0 = ')
        into.append(repr(self.value_0))
        
        return into
    
    
    @copy_docs(AssertionConditionalBase.render_failure_message_parts_into)
    def render_failure_message_parts_into(self, failure_message_parts):
        AssertionConditionalBase.render_failure_message_parts_into(self, failure_message_parts)
        _render_parameters_representation_into('parameter', self.value_0, failure_message_parts)
        return failure_message_parts


class AssertionConditionalBase2Value(AssertionConditionalBase1Value):
    """
    Base class for executing a two value assertion.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        The value to call the condition on.
    value_1 : `object`
        The value to call the condition with.
    """
    __slots__ = ('value_1',)
    
    def __new__(cls, value_0, value_1, *, reverse = False):
        """
        Asserts whether the two values are equal. Fails the test if not.
        
        Parameters
        ----------
        value_0 : `object`
            First value to assert equality with.
        value_1 : `object`
            The second value to assert equality with.
        reverse : `bool` = `False`, Optional (Keyword only)
            Whether the condition should be reversed.
        """
        self = AssertionConditionalBase1Value.__new__(cls, value_0, reverse = reverse)
        self.value_1 = value_1
        return self
    
    
    @copy_docs(AssertionConditionalBase1Value._build_repr_parts_into)
    def _build_repr_parts_into(self, into):
        into = AssertionConditionalBase1Value._build_repr_parts_into(self, into)
        
        into.append(', value_1 = ')
        into.append(repr(self.value_1))
        
        return into
    
    
    @copy_docs(AssertionConditionalBase1Value._render_operation_representation_into)
    def _render_operation_representation_into(self, into):
        AssertionConditionalBase._render_operation_representation_into(self, into)
        into.append(' as "parameter_0 ')
        into.append(self._get_operation_representation())
        into.append(' parameter_1"')
        return into
    
    
    @copy_docs(AssertionConditionalBase1Value.render_failure_message_parts_into)
    def render_failure_message_parts_into(self, failure_message_parts):
        AssertionConditionalBase.render_failure_message_parts_into(self, failure_message_parts)
        _render_parameters_representation_into('parameter_0', self.value_0, failure_message_parts)
        _render_parameters_representation_into('parameter_1', self.value_1, failure_message_parts)
        return failure_message_parts



from .call_state import *
from .handle import *
from .parameter_checking import *
from .parameter_mismatch import *
from .result_state import *


__all__ = (
    *call_state.__all__,
    *handle.__all__,
    *parameter_checking.__all__,
    *parameter_mismatch.__all__,
    *result_state.__all__,
)



__all__ = ('WrapperEnvironment',)

from scarletio import copy_docs

from ..environment.configuration import _check_environment_type

from .wrapper_base import WrapperBase


class WrapperEnvironment(WrapperBase):
    """
    Uses a specific environment for the wrapped test.

    Attributes
    ----------
    wrapped : `None`, `object`
        The wrapped test.
    environment : ``DefaultEnvironment``
        Environment to use while running the test.
    """
    __slots__ = ('environment',)
    
    def __new__(cls, environment):
        """
        Creates a environment wrapper.
        
        Parameters
        ----------
        environment : ``DefaultEnvironment``
            Environment to use while running the test.
        
        Raises
        ------
        TypeError
            - If `environment`'s type is incorrect.
        """
        _check_environment_type(environment)
        
        self = WrapperBase.__new__(cls)
        self.environment = environment
        return self
    
    
    @copy_docs(WrapperBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} environment = {self.environment!r}'
    
    
    @copy_docs(WrapperBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.environment != other.environment:
            return False
        
        return True
    
    
    @copy_docs(WrapperBase.__hash__)
    def __hash__(self):
        return hash(self.environment)
    
    
    @copy_docs(WrapperBase.is_ignored_when_testing)
    def is_ignored_when_testing(self):
        return True
    
    
    @copy_docs(WrapperBase.iter_environments)
    def iter_environments(self):
        yield self.environment



__all__ = ('AssertionValueEvaluationFalse',)

from scarletio import copy_docs

from .assertion_conditional_base import AssertionConditionalBase1Value


class AssertionValueEvaluationFalse(AssertionConditionalBase1Value):
    """
    Asserts whether the value evaluates to `False`.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        First value to assert.
    """
    __slots__ = ()
    
    @copy_docs(AssertionConditionalBase1Value.invoke_condition)
    def invoke_condition(self):
        return not self.value_0
    
    
    @copy_docs(AssertionConditionalBase1Value._get_operation_representation)
    def _get_operation_representation(self):
        return 'not bool(...)'
    
    
    @copy_docs(AssertionConditionalBase1Value._render_operation_representation_into)
    def _render_operation_representation_into(self, into):
        AssertionConditionalBase1Value._render_operation_representation_into(self, into)
        into.append(' as "not bool(parameter)"')
        return into



__all__ = ('AssertionSubtype',)

from scarletio import copy_docs

from ..helpers.un_nesting import un_nest_types

from .assertion_conditional_base import (
    AssertionConditionalBase, AssertionConditionalBase2Value, _render_parameters_representation_into,
    _render_types_parameter_representation_into
)


class AssertionSubtype(AssertionConditionalBase2Value):
    """
    Asserts whether the first object is instance of the second one.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        First value to assert subtype with.
    value_1 : `object`
        The second value to assert subtype with.
    nullable : `bool`
        Whether `value` is accepted even if given as `None`.
    """
    __slots__ = ('nullable',)
    
    def __new__(cls, value, accepted_type, *accepted_types, reverse = False, nullable = False):
        """
        Creates a new instance assertion.
        
        Parameters
        ----------
        value : `object`
            Object to check.
        accepted_type : `type`
            Type to check.
        *accepted_types : `tuple<type>`
            Additional types to check.
        reverse : `bool` = `False`, Optional (Keyword only)
            Whether the condition should be reversed.
        nullable : `bool` = `False`, Optional (Keyword only)
            Whether `value` is accepted even if given as `None`.
        
        Raises
        ------
        TypeError
            - If `accepted_type` is not `type` instance.
        ValueError
            - No types given.
        """
        accepted_types = un_nest_types((accepted_type, *accepted_types))
        if not accepted_types:
            raise ValueError(
                'At least 1 type is required.'
            )
        
        self = AssertionConditionalBase2Value.__new__(cls, value, accepted_types, reverse = reverse)
        self.nullable = nullable
        return self
    
    
    @copy_docs(AssertionConditionalBase2Value.invoke_condition)
    def invoke_condition(self): 
        value = self.value_0
        accepted_types = self.value_1
        
        if self.nullable and value is None:
            return True
        
        return isinstance(value, type) and any(issubclass(value, accepted_type) for accepted_type in accepted_types)
    
    
    @copy_docs(AssertionConditionalBase2Value._get_operation_representation)
    def _get_operation_representation(self):
        return 'subtype'
    
    
    @copy_docs(AssertionConditionalBase2Value._render_operation_representation_into)
    def _render_operation_representation_into(self, into):
        AssertionConditionalBase._render_operation_representation_into(self, into)
        into.append(' as "')
        if self.nullable:
            into.append('value is None or ')
        into.append('isinstance(value, type) and issubclass(value, expected_types)"')
        return into

    
    @copy_docs(AssertionConditionalBase2Value._build_repr_parts_into)
    def _build_repr_parts_into(self, into):
        into = AssertionConditionalBase2Value._build_repr_parts_into(self, into)
        
        nullable = self.nullable
        if nullable:
            into.append(', nullable = ')
            into.append(repr(nullable))
        
        return into
    
    
    @copy_docs(AssertionConditionalBase2Value.render_failure_message_parts_into)
    def render_failure_message_parts_into(self, failure_message_parts):
        AssertionConditionalBase.render_failure_message_parts_into(self, failure_message_parts)
        _render_parameters_representation_into('value', self.value_0, failure_message_parts)
        _render_types_parameter_representation_into('expected_types', self.value_1, failure_message_parts)
        return failure_message_parts



__all__ = ()

import sys
from os import get_terminal_size

from scarletio import RichAttributeErrorBaseType


DEFAULT_BREAK_LINE_LENGTH = 60


class OutputWriter(RichAttributeErrorBaseType):
    """
    Test output writer.
    
    Attributes
    ----------
    _last_chunk_break_line : `bool`
        Whether the last written line was a break line.
    _last_write_ended_with_linebreak : `bool`
        Whether the last write ended with a linebreak.
    file : `io-like`
        object writable object.
    """
    def __new__(cls, file = None):
        """
        Creates a new test output writer.
        
        Parameters
        ----------
        file : `None`, `io-like` = `None`, Optional
            Writable object. If not given defaults to `sys.stdout`.
        """
        if file is None:
            file = sys.stdout
        
        self = object.__new__(cls)
        self._last_chunk_break_line = False
        self._last_write_ended_with_linebreak = False
        self.file = file
        return self
    
    
    def write_break_line(self, character = '='):
        """
        Writes a break line.
        
        Parameters
        ----------
        character : `str` = `'='`, Optional
            The character to use.
        
        Returns
        -------
        written : `bool`
        """
        if self._last_chunk_break_line:
            return False
        
        if not self._last_write_ended_with_linebreak:
            self.file.write('\n')
        
        try:
            terminal_size = get_terminal_size()
        except OSError:
            break_line_length = DEFAULT_BREAK_LINE_LENGTH
        else:
            break_line_length = terminal_size.columns
        
        self.file.write(character * break_line_length)
        
        self._last_write_ended_with_linebreak = False
        self._last_chunk_break_line = True
        
        return True
    
    
    def write_line(self, string):
        """
        Writes the given line into the file.
        
        Returns
        -------
        written : `bool`
        """
        if not string:
            return False
        
        if (not self._last_write_ended_with_linebreak) and (not string.startswith('\n')):
            self.file.write('\n')
        
        self.file.write(string)
        
        self._last_write_ended_with_linebreak = string.endswith('\n')
        self._last_chunk_break_line = False
        
        return True
    
    
    def write(self, string):
        """
        Writes the given line into the file.
        
        Returns
        -------
        written : `bool`
        """
        if not string:
            return False
        
        if self._last_chunk_break_line:
            self.file.write('\n')
        
        self.file.write(string)
        
        self._last_write_ended_with_linebreak = False
        self._last_chunk_break_line = False
        
        return True
    
    
    def end_line(self):
        """
        Ends the current line.
        
        Returns
        -------
        written : `bool`
        """
        if self._last_write_ended_with_linebreak:
            written = False
        
        else:
            self.file.write('\n')
            written = True
        
        self._last_write_ended_with_linebreak = True
        self._last_chunk_break_line = False
        
        return written
    
    
    def __del__(self):
        """Clears up the file of the output writer"""
        
        try:
            file = self.file
            
            flusher = getattr(file, 'flush', None)
            if (flusher is not None):
                try:
                    flusher()
                except NotImplementedError:
                    pass
            
            closer = getattr(file, 'close', None)
            if (closer is not None):
                try:
                    closer()
                except NotImplementedError:
                    pass
        
        finally:
            # Clear up self even if exception occurs.
            self = None



__all__ = ()


def hash_object(object_):
    """
    Returns the hash value of the given object even if it not defines `__hash__`.
    
    Parameters
    ----------
    object_ : `object`
        The object to get it's hash value of.
    
    Returns
    -------
    hash_value : `int`
    """
    try:
        hash_value = hash(object_)
    except (TypeError, NotImplementedError):
        hash_value = object.__hash__(object_)
    
    return hash_value


def try_hash_method(object_):
    """
    Tries different hash methods on the given object based on it's type.
    
    Parameters
    ----------
    object_ : `object`
        The object to get it's hash value of.
    
    Returns
    -------
    hash_value : `int`
    """
    if isinstance(object_, tuple):
        return hash_tuple(object_)
    
    if isinstance(object_, list):
        return hash_list(object_)
    
    if isinstance(object_, dict):
        return hash_dict(object_)
    
    if isinstance(object_, set):
        return hash_set(object_)
    
    return hash_object(object_)


def hash_tuple(tuple_):
    """
    Returns the hash value of the given tuple.
    
    Parameters
    ----------
    tuple_ : `tuple` of `object`
        The tuple to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = len(tuple_) << 4
    for element in tuple_:
        hash_value ^= try_hash_method(element)
    
    return hash_value


def hash_list(list_):
    """
    Returns the hash value of the given list.
    
    Parameters
    ----------
    list_ : `list` of `object`
        The list to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = len(list_) << 8
    for element in list_:
        hash_value ^= try_hash_method(element)
    
    return hash_value


def hash_dict(dict_):
    """
    Returns the hash value of the given dict.
    
    Parameters
    ----------
    dict_ : `dict` of `object`
        The dict to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = len(dict_) << 12
    
    for key, value in dict_.items():
        hash_value ^= try_hash_method(key) & try_hash_method(value)
    
    return hash_value


def hash_set(set_):
    """
    Returns the hash value of the given set.
    
    Parameters
    ----------
    set_ : `set` of `object`
        The set to hash.
    
    Returns
    -------
    hash_value : `int`
    """
    hash_value = len(set_) << 16
    for element in set_:
        hash_value ^= try_hash_method(element)
    
    return hash_value



__all__ = ()

IDENTIFIER_NONE = 0
IDENTIFIER_TESTING_START = 1
IDENTIFIER_TESTING_END = 2
IDENTIFIER_FILE_REGISTRATION = 3
IDENTIFIER_FILE_REGISTRATION_DONE = 4
IDENTIFIER_FILE_LOAD_DONE = 5
IDENTIFIER_FILE_TESTING_DONE = 6
IDENTIFIER_TEST_DONE = 7
IDENTIFIER_SOURCE_LOAD_FAILURE = 8



__all__ = ()

import sys

from scarletio import create_ansi_format_code, set_docs


STYLE_RESET = create_ansi_format_code()


def style_text(text, format_code):
    """
    Styles the given text.
    
    Parameters
    ----------
    text : `str`
        Text to render with the given style.
    format_code : `str`
        Format code to style the text with.
    
    Returns
    -------
    text : `str`
    """
    return ''.join(style_text_into([], text, format_code))


if sys.platform == 'win32':
    def style_text_into(into, text, format_code):
        into.append(text)
        return into
    
    def style_text_block(into, format_code):
        yield

else:
    def style_text_into(into, text, format_code):
        into.append(format_code)
        into.append(text)
        into.append(STYLE_RESET)
        return into
    
    def style_text_block(into, format_code):
        into.append(format_code)
        yield
        into.append(STYLE_RESET)


set_docs(
    style_text_into,
    """
    Renders the given text with the style parts into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to add the new parts into.
    text : `str`
        Text to render with the given style.
    format_code : `str`
        Format code to style the text with.
    
    Returns
    -------
    into : `list` of `str`
    """
)


set_docs(
    style_text_block,
    """
    Renders the style into the given parts. Can be used as a block.
    
    ```py
    for _ in style_text_into(into, format_code):
        into.append(text_part)
        ...
    ```
    
    This function is an iterable generator.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to add the new parts into.
    format_code : `str`
        Format code to style the text with.
    """
)



__all__ = ('ReportFailureRaising',)

from scarletio import copy_docs

from .base import ReportBase


class ReportFailureRaising(ReportBase):
    """
    Test failure representing a failed raising exception check.
    
    Attributes
    ----------
    accept_subtypes : `bool`
        Whether exception subclasses were allowed.
    expected_exceptions : `None`, `set` of `BaseException`
        Expected raised exceptions.
    exception_received : `None`, `BaseException`
        The received exception.
    """
    __slots__ = ('accept_subtypes', 'expected_exceptions', 'exception_received',)
    
    def __new__(cls, expected_exceptions, exception_received, accept_subtypes):
        """
        Creates a new raising test failure.
        
        Parameters
        ----------
        expected_exceptions : `None`, `set` of `BaseException`
            Expected raised exceptions.
        exception_received : `None`, `BaseException`
            The received exception.
        accept_subtypes : `bool`
            Whether exception subclasses were allowed.
        """
        self = object.__new__(cls)
        self.expected_exceptions = expected_exceptions
        self.exception_received = exception_received
        self.accept_subtypes = accept_subtypes
        return self
    
    
    @copy_docs(ReportBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' expected_exceptions = ')
        repr_parts.append(repr(self.expected_exceptions))
        
        repr_parts.append(', received_exception = ')
        repr_parts.append(repr(self.exception_received))
        
        accept_subtypes = self.accept_subtypes
        if accept_subtypes:
            repr_parts.append(', accept_subtypes = ')
            repr_parts.append(repr(accept_subtypes))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ReportBase.is_failure)
    def is_failure(self):
        return True



__all__ = ('ReportFailureReturning',)

from scarletio import copy_docs

from .base import ReportBase


class ReportFailureReturning(ReportBase):
    """
    Test failure representing a bad return value.
    
    Attributes
    ----------
    expected_value : `None`, `object`
        The expected returned value.
    received_value : `None`, `object`
        The received returned value.
    """
    __slots__ = ('expected_value', 'received_value',)
    
    def __new__(cls, expected_value, received_value):
        """
        Creates a new test failure representing a bad return value.
        
        Parameters
        ----------
        expected_value : `None`, `object`
            The expected returned value.
        received_value : `None`, `object`
            The received returned value.
        """
        self = object.__new__(cls)
        self.expected_value = expected_value
        self.received_value = received_value
        return self
    
    
    @copy_docs(ReportBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' expected_value = ')
        repr_parts.append(repr(self.expected_value))
        
        repr_parts.append(', received_value = ')
        repr_parts.append(repr(self.received_value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ReportBase.is_failure)
    def is_failure(self):
        return True



from .assertions import *
from .contexts import *
from .environment import *
from .event_handling import *
from .events import *
from .file import *
from .handling import *
from .helpers import *
from .mocking import *
from .result import *
from .runner import *
from .utils import *
from .wrappers import *

from .run_state import *
from .test_case import *


__all__ = (
    *assertions.__all__,
    *contexts.__all__,
    *environment.__all__,
    *event_handling.__all__,
    *events.__all__,
    *file.__all__,
    *handling.__all__,
    *helpers.__all__,
    *mocking.__all__,
    *result.__all__,
    *runner.__all__,
    *utils.__all__,
    *wrappers.__all__,
    
    *run_state.__all__,
    *test_case.__all__,
)



from .load_failure_rendering import *
from .result_modifier_parameters import *
from .result_rendering import *
from .writers import *


__all__ = (
    *load_failure_rendering.__all__,
    *result_modifier_parameters.__all__,
    *result_rendering.__all__,
    *writers.__all__,
)



__all__ = ()

from os import get_terminal_size

from scarletio import DEFAULT_ANSI_HIGHLIGHTER, render_exception_into

from ...environment.default import __file__ as VAMPYTEST_ENVIRONMENT_DEFAULT_FILE_PATH
from ...environment.scarletio_coroutine import __file__ as VAMPYTEST_ENVIRONMENT_SCARLETIO_COROUTINE_FILE_PATH

from ..colors import COLOR_FAIL, COLOR_PATH, COLOR_SKIP
from ..default_output_writer import DEFAULT_BREAK_LINE_LENGTH
from ..text_styling import style_text_block, style_text_into


def create_break(character):
    """
    Creates a break with the given character.
    
    Parameters
    ----------
    character : `str`
        The character to create the break from.
    
    Returns
    -------
    break : `str`
    """
    try:
        terminal_size = get_terminal_size()
    except OSError:
        break_line_length = DEFAULT_BREAK_LINE_LENGTH
    else:
        break_line_length = terminal_size.columns

    return character * break_line_length


def render_route_parts_into(into, result):
    """
    Adds route parts into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    result : ``Result``
        Test result to get its path of.
    
    Returns
    -------
    into : `list` of `str`
    """
    case = result.handle.case
    
    into.append(case.import_route)
    into.append('.')
    into.append(case.name)
    
    return into


def render_documentation_into(into, result):
    """
    Adds the documentation into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    result : ``Result``
        Test result to get its path of.
    
    Returns
    -------
    into : `list` of `str`
    """
    documentation = result.handle.get_test_documentation()
    if (documentation is not None):
        into.append('\n')
        into.append(documentation)
    
    return into


def render_test_position_into(into, title, title_format_code, result):
    """
    Renders the test's position into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    title : `str`
        Title to color.
    title_format_code : `str`
        Format code to format the title.
    result : ``Result``
        Test result to get its path of.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = style_text_into(into, title, title_format_code)
    into.append(' at: ')
    for _ in style_text_block(into, COLOR_PATH):
        into = render_route_parts_into(into, result)
    
    render_documentation_into(into, result)
    return into


def render_parameters_into(into, positional_parameters, keyword_parameters):
    """
    Renders the input parameters of to the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    positional_parameters : `None`, `list<object>`
        Positional parameters passed to the test.
    keyword_parameters : `None`, `dict<str, object>`
        Keyword parameters passed to the test.
    
    Returns
    -------
    into : `list` of `str`
    """
    if (positional_parameters is not None):
        for parameter_value in positional_parameters:
            into.append('\n    ')
            into.append(repr(parameter_value))
    
    
    if (keyword_parameters is not None):
        for (parameter_name, parameter_value) in keyword_parameters.items():
            into.append('\n    ')
            into.append(parameter_name)
            into.append(' = ')
            into.append(repr(parameter_value))
    
    if (positional_parameters is None) and (keyword_parameters is None):
        into.append(' N/A')
    
    return into


def maybe_render_parameters_section_into(into, result):
    """
    Renders the parameters section into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    result : ``Result``
        Test result.
    
    Returns
    -------
    into : `list` of `str`
    """
    final_call_state = result.handle.final_call_state
    if (final_call_state is not None) and final_call_state:
        into.append('\n\nParameters:')
        into = render_parameters_into(into, final_call_state.positional_parameters, final_call_state.keyword_parameters)
    
    return into


def maybe_render_output_into(into, result):
    """
    Renders the captured output section into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    result : ``Result``
        Test result.
    
    Returns
    -------
    into : `list` of `str`
    """
    report = result.get_output_report()
    if (report is not None):
        if into and (not into[-1].endswith('\n')):
            into.append('\n')
        
        into.append('\nCaptured output while running the test:\n')
        into.append(create_break('-'))
        into.append('\n')
        into.append(report.output)
    
    return into


def render_report_reversed_into(into, result):
    """
    Creates a reversed failure message for the given test handle.
    
    Parameters
    ----------
    into : `list` of `str`
        A list to put the string parts into.
    result : ``Result``
        Test result.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Reversed test passed', COLOR_FAIL, result)
    
    final_call_state = result.handle.final_call_state
    if (final_call_state is not None) and final_call_state:
        into.append('\nParameters: ')
        into = render_parameters_into(into, final_call_state.positional_parameters, final_call_state.keyword_parameters)
    
    return into


def render_wrapper_conflict_into(into, result):
    """
    Renders wrapper conflict.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The result failing with wrapper conflict.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Wrapper conflict', COLOR_FAIL, result)
    
    wrapper_conflict = result.wrapper_conflict
    reason = wrapper_conflict.reason
    if (reason is not None):
        into.append('\nReason: ')
        into = style_text_into(into, reason, COLOR_FAIL)
    
    into.append('\nBetween wrapper(s):\n- ')
    into.append(repr(wrapper_conflict.wrapper_0))
    
    wrapper_1 = wrapper_conflict.wrapper_1
    if (wrapper_1 is not None):
        into.append('\n- ')
        into.append(repr(wrapper_1))
    
    return into


def ignore_invoke_test_frame(frame):
    """
    Ignores the frame where the test is called from.
    
    Parameters
    ----------
    frame : ``FrameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.file_name
    name = frame.name
    line = frame.line
    
    if file_name == VAMPYTEST_ENVIRONMENT_DEFAULT_FILE_PATH:
        if name == 'run':
            if line == 'returned_value = test(*positional_parameters, **keyword_parameters)':
                should_show_frame = False
    
    elif file_name == VAMPYTEST_ENVIRONMENT_SCARLETIO_COROUTINE_FILE_PATH:
        if name == '_run_async':
            if line == 'returned_value = await test(*positional_parameters, **keyword_parameters)':
                should_show_frame = False
    
    return should_show_frame


def render_report_asserting_into(into, result, report):
    """
    Renders assertion failure report.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The ran test.
    report : ``ReportFailureAsserting``
        Report containing the failure information.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Assertion failed', COLOR_FAIL, result)
    maybe_render_parameters_section_into(into, result)
    into.append('\n')
    into = report.assertion_exception.render_failure_message_parts_into(into)
    into = maybe_render_output_into(into, result)
    return into


def render_report_raising_into(into, result, report):
    """
    Renders raising failure report.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The ran test.
    report : ``ReportFailureRaising``
        Report containing the failure information.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Unexpected exception', COLOR_FAIL, result)
    maybe_render_parameters_section_into(into, result)
    into.append('\n')
    
    expected_exceptions = report.expected_exceptions
    if expected_exceptions is not None:
        into.append('\nExpected exceptions: ')
        
        exception_added = False
        
        for expected_exception in expected_exceptions:
            if exception_added:
                into.append(', ')
            else:
                exception_added = True
            
            if isinstance(expected_exception, type):
                expected_exception_representation = expected_exception.__name__
            else:
                expected_exception_representation = repr(expected_exception)
            
            into.append(expected_exception_representation)
        
        into.append('\nAccept sub-types: ')
        into.append('true' if report.accept_subtypes else 'false')
    
    
    exception_received = report.exception_received
    if (exception_received is not None):
        into.append('\n')
        into.append(create_break('-'))
        into.append('\n')
        
        render_exception_into(
            exception_received,
            into,
            filter = ignore_invoke_test_frame,
            highlighter = DEFAULT_ANSI_HIGHLIGHTER
        )
    
    into = maybe_render_output_into(into, result)
    return into


def render_report_returning_into(into, result, report):
    """
    Renders returning failure report.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The ran test.
    report : ``ReportFailureReturning``
        Report containing the failure information.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Unexpected return', COLOR_FAIL, result)
    maybe_render_parameters_section_into(into, result)
    
    into.append('\n\nExpected return: ')
    into.append(repr(report.expected_value))
    
    into.append('\nReceived return: ' )
    into.append(repr(report.received_value))
    
    into = maybe_render_output_into(into, result)
    return into


def render_report_output_into(into, result, report):
    """
    Renders output report.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The ran test.
    report : ``ReportOutput``
        Report containing the output.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Output captured', COLOR_SKIP, result)
    maybe_render_parameters_section_into(into, result)
    
    into.append('\n')
    into.append(create_break('-'))
    into.append('\n')
    into.append(report.output)
    
    return into


def render_function_parameters_into(into, parameters):
    """
    Renders a function's parameters.
    
    Parameters
    ----------
    into : `list<str>`
        Container to render into.
    parameters : `list<Parameter>`
        The parameter to render.
    
    Returns
    -------
    into : `list<str>`
    """
    previous = None
    
    for parameter in parameters:
        if previous is not None:
            if previous.is_positional_only():
                if parameter.is_keyword_only():
                    into.append('\n    /\n    *')
                elif (not parameter.is_positional_only()):
                    into.append('\n    /')
            
            elif previous.is_positional():
                if parameter.is_keyword_only():
                    into.append('\n    *')
        
        into.append('\n    ')
        
        # * or ** ?
        if parameter.is_args():
            into.append('*')
        elif parameter.is_kwargs():
            into.append('**')
        
        # name
        into.append(parameter.name)
        
        # default
        if parameter.has_default:
            into.append(' = ')
            into.append(repr(parameter.default))
        
        previous = parameter
    
    return into


def render_unsatisfied_parameters_into(issue_parts, unsatisfied_parameters):
    """
    Renders the unsatisfied parameters.
    
    Parameters
    ----------
    into : `list<str>`
        Container to render into.
    parameters : `list<Parameter>`
        The parameter to render.
    
    Returns
    -------
    into : `list<str>`
    """
    for parameter in unsatisfied_parameters:
        issue_parts.append('\n    ')
        issue_parts.append(parameter.name)
    
    return issue_parts


def render_extra_parameters_into(issue_parts, extra_positional_parameters, extra_keyword_parameters):
    if (extra_positional_parameters is not None):
        for parameter_value in extra_positional_parameters:
            issue_parts.append('\n    ')
            issue_parts.append(repr(parameter_value))
    
    if (extra_keyword_parameters is not None):
        for parameter_name, parameter_value in extra_keyword_parameters:
            issue_parts.append('\n    ')
            issue_parts.append(parameter_name)
            issue_parts.append(' = ')
            issue_parts.append(repr(parameter_value))
    
    return issue_parts


def render_report_parameter_mismatch_into(into, result, report):
    """
    Renders parameter mismatch failure report.
    
    Parameter
    ---------
    into : `list` of `str`
        List to render into.
    result : ``Result``
        The ran test.
    report : ``ReportFailureParameterMismatch``
        Report containing the failure information.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_test_position_into(into, 'Parameter mismatch', COLOR_FAIL, result)
    parameter_mismatch = report.parameter_mismatch
    
    into.append('\n\nFunction parameters:')
    into = render_function_parameters_into(into, parameter_mismatch.parameters)
    
    into.append('\n\nGiven Parameters:')
    into = render_parameters_into(into, parameter_mismatch.positional_parameters, parameter_mismatch.keyword_parameters)
    
    unsatisfied_parameters = parameter_mismatch.unsatisfied_parameters
    if (unsatisfied_parameters is not None):
        into.append('\n\nUnsatisfied function parameters:')
        into = render_unsatisfied_parameters_into(into, unsatisfied_parameters)
    
    extra_positional_parameters = parameter_mismatch.extra_positional_parameters
    extra_keyword_parameters = parameter_mismatch.extra_keyword_parameters
    if (extra_positional_parameters is not None) or (extra_keyword_parameters is not None):
        into.append('\n\nExtra parameters:')
        into = render_parameters_into(into, extra_positional_parameters, extra_keyword_parameters)
    
    return into



__all__ = ('execute_from_terminal',)

import sys
from os import getcwd as get_current_working_directory

from ..core import run_tests_in, shutdown_environments

from .source_lookup import get_source_and_target


def build_error_message(errors, parameters):
    """
    Builds error message for the case when test source could not be identified.
    
    Parameters
    ----------
    errors : `None | list<(str, str)>`
        Found errors while trying to resolve source. In a `path - message` relation.
    parameters : `list<str>`
        Parameters the test would be called with.
    
    Returns
    -------
    error_message : `str`
    """
    error_message_parts = []
    
    error_message_parts.append('Could not detect testing directory.\n')
    
    error_message_parts.append('Working directory: ')
    error_message_parts.append(get_current_working_directory())
    error_message_parts.append('\n')
    
    error_message_parts.append('Parameters: ')
    if parameters:
        for parameter in parameters:
            error_message_parts.append('\n- "')
            error_message_parts.append(parameter.replace('"', '\\"'))
            error_message_parts.append('"')
    
    else:
        error_message_parts.append('*none*')
    error_message_parts.append('\n')
    
    if errors is not None:
        error_message_parts.append('\nErrors while detecting source(s):\n')
        
        for index, (path, error) in enumerate(errors, 1):
            index_string = str(index)
            error_message_parts.append(index_string)
            error_message_parts.append('. ')
            error_message_parts.append('Path: ')
            error_message_parts.append(path)
            error_message_parts.append(' ' * (2 + len(index_string)))
            error_message_parts.append('\n')
            
            error_message_parts.append('Error: ')
            error_message_parts.append(error)
            error_message_parts.append('\n')
    
    return ''.join(error_message_parts)


def execute_from_terminal(parameters):
    """
    Executes vampytest from terminal.
    
    Parameters
    ----------
    parameters : `list` of `str`
        Command line parameters.
    
    Returns
    -------
    success : `bool`
        This does not mean all tests passed.
    """
    source_directory, sources, test_collection_route, errors, index = get_source_and_target(parameters, 0)
    
    if (source_directory is None) or (sources is None):
        sys.stderr.write(build_error_message(errors, parameters))
        return False
    
    
    try:
        run_tests_in(source_directory, sources, test_collection_route)
    finally:
        shutdown_environments()
    
    return True



__all__ = ('AssertionEquals',)

from scarletio import copy_docs

from .assertion_conditional_base import AssertionConditionalBase2Value


class AssertionEquals(AssertionConditionalBase2Value):
    """
    Asserts equality.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    exception : `None`, `BaseException`
        Exception raised by the condition if any.
    reverse : `bool`
        Whether the condition should be reversed.
    value_0 : `object`
        First value to assert equality with.
    value_1 : `object`
        The second value to assert equality with.
    """
    __slots__ = ()
    
    @copy_docs(AssertionConditionalBase2Value.invoke_condition)
    def invoke_condition(self):
        return self.value_0 == self.value_1
    
    
    @copy_docs(AssertionConditionalBase2Value._get_operation_representation)
    def _get_operation_representation(self):
        return '=='



__all__ = ('AssertionException', )

from scarletio import export, render_frames_into
from scarletio.utils.trace.frame_proxy import get_exception_frames

from ..environment.default import __file__ as VAMPYTEST_ENVIRONMENT_DEFAULT_FILE_PATH
from ..environment.scarletio_coroutine import __file__ as VAMPYTEST_ENVIRONMENT_SCARLETIO_COROUTINE_FILE_PATH

from .assertion_conditional_base import __file__ as VAMPYTEST_ASSERTION_CONDITION_BASE_FILE_PATH
from .assertion_instance import __file__ as VAMPYTEST_ASSERTION_INSTANCE_FILE_PATH
from .assertion_raising import __file__ as VAMPYTEST_ASSERTION_RAISING_FILE_PATH
from .assertion_subtype import __file__ as VAMPYTEST_ASSERTION_SUBTYPE_FILE_PATH


def _ignore_assertion_frames(frame):
    """
    Ignores the frame where assertions are done.
    
    Parameters
    ----------
    frame : ``frameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.file_name
    name = frame.name
    line = frame.line
    
    if file_name == VAMPYTEST_ASSERTION_CONDITION_BASE_FILE_PATH:
        if name == '__call__':
            if line == 'return type_.__new__(type_, *positional_parameters, **keyword_parameter).invoke()':
                should_show_frame = False
        
        elif name == 'invoke':
            if line == 'raise AssertionException(self)':
                should_show_frame = False
    
    elif file_name == VAMPYTEST_ASSERTION_INSTANCE_FILE_PATH:
        if name == '__new__':
            if line == 'return self.invoke()':
                should_show_frame = False

    elif file_name == VAMPYTEST_ASSERTION_SUBTYPE_FILE_PATH:
        if name == '__new__':
            if line == 'return self.invoke()':
                should_show_frame = False
    
    elif file_name == VAMPYTEST_ASSERTION_RAISING_FILE_PATH:
        if name == '__exit__':
            if line == 'raise AssertionException(self)':
                should_show_frame = False
    
    elif file_name == VAMPYTEST_ENVIRONMENT_DEFAULT_FILE_PATH:
        if name == 'run':
            if line == 'returned_value = test(*positional_parameters, **keyword_parameters)':
                should_show_frame = False
    
    elif file_name == VAMPYTEST_ENVIRONMENT_SCARLETIO_COROUTINE_FILE_PATH:
        if name == '_run_async':
            if line == 'returned_value = await test(*positional_parameters, **keyword_parameters)':
                should_show_frame = False
    
    
    return should_show_frame


@export
class AssertionException(BaseException):
    """
    Raised when an exception fails.
    
    Attributes
    ----------
    assertion : ``AssertionBase``
        The failed assertion.
    """
    __slots__ = ('assertion',)
    __init__ = object.__init__
    
    def __new__(cls, assertion):
        """
        Creates a new condition exception.
        
        Parameters
        ----------
        assertion : ``AssertionBase``
            The failed assertion.
        """
        self = BaseException.__new__(cls, assertion)
        self.assertion = assertion
        return self
    
    
    def render_failure_message_parts_into(self, failure_message_parts):
        """
        Renders the exception into the given list.
        
        Parameters
        ----------
        failure_message_parts : `list` of `str`
            A list to put the rendered strings into.
        
        Returns
        -------
        failure_message_parts : `list` of `str`
        """
        frames = get_exception_frames(self)
        failure_message_parts.append('\n')
        render_frames_into(frames, failure_message_parts, filter = _ignore_assertion_frames)
        failure_message_parts.append('\n')
        self.assertion.render_failure_message_parts_into(failure_message_parts)
        return failure_message_parts



__all__ = ('WrapperBase', )

from types import FunctionType

from scarletio import RichAttributeErrorBaseType, include

from ..helpers.hashing import hash_object


WrapperChainer = include('WrapperChainer')


class WrapperBase(RichAttributeErrorBaseType):
    """
    Base class for test wrappers defining shared functionality.
    
    Attributes
    ----------
    wrapped : `None`, ``WrapperBase``, `FunctionType`
        The wrapped test.
    """
    __slots__ = ('wrapped', )
    
    def __new__(cls):
        """
        Creates a new test wrapper.
        """
        self = object.__new__(cls)
        self.wrapped = None
        return self
    
    
    def __call__(self, to_wrap):
        """
        Calls the wrapper.
        
        Parameters
        ----------
        to_wrap : `object`
            The object to wrap. It can be either a function or an another wrapper.
        
        Returns
        -------
        to_wrap : ``WrapperBase``
            The same or a newly created wrapper.
        
        Raises
        ------
        RuntimeError
            Wrapper already called.
        """
        if to_wrap is None:
            raise RuntimeError(f'Cannot wrap `None`; self = {self!r}.')
        
        
        wrapped = self.wrapped
        
        if isinstance(to_wrap, WrapperBase):
            
            if (wrapped is not None) and (to_wrap.wrapped is not None):
                raise RuntimeError(
                    f'Both self and other wrappers are already wrapped; self = {self!r}; other = {to_wrap!r}.'
                )
            
            if isinstance(to_wrap, WrapperChainer):
                to_wrap.append(self)
                
                if (wrapped is not None):
                    to_wrap(wrapped)
                
                return to_wrap
            
            if wrapped is None:
                wrapped = to_wrap.wrapped
            
            wrapper_chainer = WrapperChainer(wrapped)
            wrapper_chainer.append(self)
            wrapper_chainer.append(to_wrap)
            return wrapper_chainer
        
        
        if (wrapped is not None):
            raise RuntimeError(f'Wrapped already wrapped; self = {self!r}, to_wrap = {to_wrap!r}.')
        
        
        if not isinstance(to_wrap, FunctionType):
            raise TypeError(
                f'Only functions and other wrappers can be wrapped; Got self = {self!r}; to_wrap = {to_wrap!r}.'
            )
        
        self.wrapped = to_wrap
        return self
    
    
    def __matmul__(self, other):
        """Returns `self(other)`"""
        return self.__call__(other)
    
    
    def __rmatmul__(self, other):
        """Returns `self(other)`"""
        return self.__call__(other)
        
    
    def _cursed_repr_builder(self):
        """
        Representation builder helper.
        
        This method is a generator.
        
        Examples
        --------
        ```
        for field_added, repr_parts in self._cursed_repr_builder():
            if not field_added:
                repr_parts.append(',')
            
            repr_parts.append(' oh no')
        
        return ''.join(repr_parts)
        ```
        """
        repr_parts = ['<', self.__class__.__name__]
        
        wrapped = self.wrapped
        if (wrapped is None):
            repr_parts.append(' wrapped = ')
            repr_parts.append(repr(wrapped))
            
            field_added = True
        else:
            field_added = False
        
        yield field_added, repr_parts
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __repr__(self):
        """Returns the wrapper's representation."""
        for field_added, repr_parts in self._cursed_repr_builder():
            pass
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the wrapper's representation."""
        wrapped = self.wrapped
        if wrapped is None:
            hash_value = 0
        else:
            hash_value = hash_object(wrapped)
        
        return hash_value
    
    
    def __eq__(self, other):
        """return whether the two wrappers are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.wrapped != other.wrapped:
            return False
        
        return True
    
    
    def do_skip(self):
        """
        Whether the test should be skipped.
        
        Returns
        -------
        do_skip : `bool`
        """
        return False
    
    
    def do_reverse(self):
        """
        Tests whether the test's result should be reversed.
        
        Returns
        -------
        do_reverse : `bool`
        """
        return False
    
    
    def check_conflicts(self):
        """
        Checks whether the wrapper has internal conflict.
        
        Returns
        ------
        wrapper_conflict: `None`, ``WrapperConflict``
        """
        pass
    
    
    def check_conflict_with(self, other):
        """
        Checks whether the wrapper has conflict with the other one.
        
        Returns
        ------
        wrapper_conflict: `None`, ``WrapperConflict``
        """
        pass
    
    
    def has_bound_test(self):
        """
        Returns whether the wrapper wraps a test.
        
        Returns
        -------
        has_bound_test : `bool`
        """
        return (self.wrapped is not None)
    
    
    def unbind_test(self):
        """
        Unbinds the wrappers test.
        
        Returns
        -------
        test : `object`
        
        Raises
        ------
        RuntimeError
            The wrapper had no test or already unbind.
        """
        wrapped = self.wrapped
        if wrapped is None:
            raise RuntimeError(f'The wrapper had no test or already unbind; self = {self!r}.')
        
        self.wrapped = None
        return wrapped
    
    
    def iter_wrappers(self):
        """
        Iterates over the wrappers encapsulated by this wrapper. This method is used to un-nest wrapper groups if
        required.
        
        This method is an iterable generator.
        
        Yields
        ------
        wrapper : ``WrapperBase``
        """
        yield self
    
    
    def is_ignored_when_testing(self):
        """
        Returns whether the wrapper can be ignore when testing, but is used instead for just pre-checks.
        
        Returns
        -------
        is_ignored_when_testing : `bool`
        """
        return True
    
    
    def is_mutually_exclusive_with(self, other):
        """
        Returns whether the wrapper is mutually exclusive with the other one.
        
        Parameters
        ----------
        other : ``WrapperBase``
        
        Returns
        -------
        is_mutually_exclusive_with : `bool`
        """
        return False
    
    
    def get_context(self, handle):
        """
        Returns a context over a test handle.
        
        Parameters
        ----------
        handle : ``Handle``
            Test handle.
        
        Returns
        -------
        context : ``ContextBase``
        """
        return None
    
    
    def iter_environments(self):
        """
        Iterates over the environments registered to the wrapper.
        
        This method is an iterable generator.
        
        Yields
        ------
        environment : ``DefaultEnvironment``
        """
        return
        yield



from .rendering_helpers import *

from .base import *
from .colors import *
from .default import *
from .default_output_writer import *
from .text_styling import *


__all__ = (
    *rendering_helpers.__all__,
    
    *base.__all__,
    *colors.__all__,
    *default.__all__,
    *default_output_writer.__all__,
    *text_styling.__all__,
)



__all__ = ('Handle',)

from reprlib import repr as short_repr

from scarletio import RichAttributeErrorBaseType, include

from ..contexts import ContextOutputCapturing

from .call_state import CallState
from .parameter_checking import check_parameter_mismatch


Result = include('Result')
AssertionException = include('AssertionException')


class Handle(RichAttributeErrorBaseType):
    """
    Handles a test.
    
    Attributes
    ----------
    case : ``TestCase``
        The parent test case.
    environments : `None`, `tuple` of ``Environment``
        Environments to use to use when testing.
    final_call_state : `None`, ``CallState``
        The final call state, with which the test is called.
    final_result_state : `None`, ``ResultState``
        Final result state after wrappers processed the original.
    original_call_state : `None`, ``CallState``
        Vanilla call state.
    original_result_state : `None`, ``ResultState``
        Result state created from the test's result.
    test : ``FunctionType``
        The test to invoke.
    wrappers : `None`, `tuple` of ``WrapperBase``
        Wrappers wrapping the test.
    """
    __slots__ = (
        'case', 'environments', 'final_call_state', 'final_result_state', 'original_call_state',
        'original_result_state', 'test', 'wrappers'
    )
    
    def __new__(cls, case, test, wrappers, environments):
        """
        Creates a new test handle.
        
        Parameters
        ----------
        case : ``TestCase``
            The parent test case.
        test : `FunctionType`
            The test to call.
        wrappers : `None`, `tuple` of ``WrapperBase``
            Wrappers wrapping the test.
        environments : `None`, `tuple` of ``Environment``
            Environments to use to use when testing.
        """
        self = object.__new__(cls)
        
        self.case = case
        self.test = test
        self.wrappers = wrappers
        self.environments = environments
        
        self.original_call_state = None
        self.final_call_state = None
        
        self.original_result_state = None
        self.final_result_state = None
        
        return self
    
    
    def __repr__(self):
        """Returns the test handle's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' test = ')
        repr_parts.append(short_repr(self.test))
        
        wrappers = self.wrappers
        if (wrappers is not None):
            repr_parts.append(', wrappers = ')
            repr_parts.append(repr(wrappers))
        
        original_call_state = self.original_call_state
        if (original_call_state is not None) and original_call_state:
            repr_parts.append(', original_call_state = ')
            repr_parts.append(repr(original_call_state))
        
        final_call_state = self.final_call_state
        if (final_call_state is not None) and final_call_state:
            repr_parts.append(', final_call_state = ')
            repr_parts.append(repr(final_call_state))
            
        original_result_state = self.original_result_state
        if (original_result_state is not None) and original_result_state:
            repr_parts.append(', original_result_state = ')
            repr_parts.append(repr(original_result_state))
        
        final_result_state = self.final_result_state
        if (final_result_state is not None) and final_result_state:
            repr_parts.append(', final_result_state = ')
            repr_parts.append(repr(final_result_state))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def _collect_contexts_into(self, contexts):
        """
        Initialises the wrappers of the test handle.
        
        Parameters
        ----------
        contexts : `list` of ``ContextBase```
            Test contexts to put the new ones into.
        """
        wrappers = self.wrappers
        if wrappers is not None:
            for wrapper in wrappers:
                context = wrapper.get_context(self)
                if (context is not None):
                    contexts.append(context)
    
    
    def _start_contexts(self, contexts):
        """
        Starts the text wrapper contexts.
        
        Parameters
        ----------
        contexts : `list` of ``ContextBase``
            Contexts to start.
        
        Returns
        -------
        test_result : `None`, ``Result``
            Result of the test.
        """
        for context in contexts:
            test_result = context.start()
            if (test_result is not None):
                return test_result
    
    
    def _enter_contexts(self, contexts):
        """
        Enters the text wrapper contexts passing them a ``CallState``, expecting them to return a new call state, or
        a test result.
        
        Parameters
        ----------
        contexts : `list` of ``ContextBase``
            Contexts to enter.
        
        Returns
        -------
        test_result : `None`, ``Result``
            Result of the test.
        """
        call_state = CallState()
        self.original_call_state = call_state
        
        try:
            for context in contexts:
                test_result, new_call_state = context.enter(call_state)
                
                if (new_call_state is not None):
                    call_state = new_call_state
                
                if (test_result is not None):
                    return test_result
        
        finally:
            self.final_call_state = call_state
    
    
    def _exit_contexts(self, contexts):
        """
        Exits the contexts passing them a ``ResultState``, expecting them to return a result state, or a result.
        
        Parameters
        ----------
        contexts : `list` of ``ContextBase``
            Contexts to exit.
        
        Returns
        -------
        test_result : `None`, ``Result``
            Result of the test.
        """
        result_state = self.original_result_state
        
        try:
            for context in reversed(contexts):
                test_result, new_result_state = context.exit(result_state)
                
                if (new_result_state is not None):
                    result_state = new_result_state
                
                if (test_result is not None):
                    return test_result
            
        finally:
            self.final_result_state = result_state
    
    
    def _close_contexts(self, contexts, result):
        """
        Exits the contexts.
        
        Parameters
        ----------
        contexts : `list` of ``ContextBase``
            Contexts to exit.
        result : `None`, ``Result˙`
            Result of the test.
        """
        for context in reversed(contexts):
            context.close(result)
    
    
    def _get_call_parameters_nullable(self):
        """
        Gets the call parameters to invoke the test with. The output elements are nullable.
        
        Returns
        -------
        positional_parameters : `None | list<object>`
            Positional parameters to call the test with.
        keyword_parameters : `None | dict<str, object>`
            Keyword parameters to call the test with.
        """
        call_state = self.final_call_state
        
        if call_state is None:
            positional_parameters = None
            keyword_parameters = None
        
        else:
            positional_parameters = call_state.positional_parameters
            keyword_parameters = call_state.keyword_parameters
        
        return positional_parameters, keyword_parameters
    
    
    def _get_call_parameters(self):
        """
        Gets call parameters to invoke the test with.
        
        Returns
        -------
        positional_parameters : `list` of `object`
            Positional parameters to call the test with.
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the test with.
        """
        call_state = self.final_call_state
        
        if call_state is None:
            positional_parameters = None
            keyword_parameters = None
        
        else:
            positional_parameters = call_state.positional_parameters
            keyword_parameters = call_state.keyword_parameters
        
        if (positional_parameters is None):
            positional_parameters = []
        
        if (keyword_parameters is None):
            keyword_parameters = {}
        
        return positional_parameters, keyword_parameters
    
    
    def _check_call_parameter_match(self):
        """
        Checks whether the call parameters match with the test's.
        
        Returns
        -------
        test_result : `None`, ``Result``
            Result of the test.
        """
        parameter_mismatch = check_parameter_mismatch(self.test, *self._get_call_parameters_nullable())
        if (parameter_mismatch is not None):
            return Result(self.case).with_handle(self).with_parameter_mismatch(parameter_mismatch)
    
    
    def _invoke_test(self, environment_manager):
        """
        Invokes the test of the test handle.
        
        Parameters
        ----------
        environment_manager : ``EnvironmentManager``
            Testing environment manager.
        """
        environment = environment_manager.get_environment_for_test(self.test)
        self.original_result_state = environment.run(self.test, *self._get_call_parameters())
    
    
    def _build_default_test_result(self):
        """
        Builds test result if non of the wrappers did before.
        
        Returns
        -------
        test_result : ``Result``
            Result of the test.
        """
        test_result = Result(self.case).with_handle(self)
        
        result_state = self.final_result_state
        if (result_state is not None) and result_state.is_raise():
            raised_exception = result_state.result
            if isinstance(raised_exception, AssertionException):
                test_result = test_result.with_assertion(raised_exception)
            else:
                test_result = test_result.with_exception(None, raised_exception, False)
        
        return test_result
    
    
    def _iter_applied_environments(self):
        """
        Iterates over the applies the environment wrappers to the test.
        
        This method is an iterable generator.
        
        Yields
        ------
        environment : ``Environment``
        """
        environments = self.environments
        if (environments is not None):
            yield from environments
    
    
    def invoke(self, environment_manager):
        """
        Invokes the test.
        
        Parameters
        ----------
        environment_manager : ``EnvironmentManager``
            Testing environment manager.
        
        Returns
        -------
        test_result : ``Result``
            Result of the test.
        """
        environment_manager = environment_manager.with_environment(*self._iter_applied_environments())
        
        contexts = [ContextOutputCapturing()]
        test_result = None
        
        try:
            self._collect_contexts_into(contexts)
            
            test_result = self._start_contexts(contexts)
            if (test_result is not None):
                return test_result
            
            test_result = self._enter_contexts(contexts)
            if (test_result is not None):
                return test_result
            
            test_result = self._check_call_parameter_match()
            if (test_result is not None):
                return test_result
            
            self._invoke_test(environment_manager)
            
            test_result = self._exit_contexts(contexts)
            if (test_result is not None):
                return test_result
            
            test_result = self._build_default_test_result()
            return test_result
        
        finally:
            self._close_contexts(contexts, test_result)
    
    
    def get_test_documentation(self):
        """
        Returns the test's documentation.
        
        Returns
        -------
        documentation : `None`, `str`
        """
        test_name = getattr(self.test, '__name__', None)
        if (test_name is None) or (not isinstance(test_name, str)):
            return None
        
        if self.case.name != test_name:
            return None
        
        raw_documentation = getattr(self.test, '__doc__', None)
        if (raw_documentation is None) or (not isinstance(raw_documentation, str)):
            return None
        
        lines = raw_documentation.split('\n')
        if not lines:
            return None
        
        for index in range(len(lines)):
            lines[index] = lines[index].rstrip()
        
        indentation_to_remove = 0
        
        while True:
            for line in lines:
                if len(line) <= indentation_to_remove:
                    continue
                
                if line[indentation_to_remove] in (' ', '\t'):
                    continue
                
                break
            
            else:
                indentation_to_remove += 1
                continue
            
            break
        
        
        for index in range(len(lines)):
            lines[index] = lines[index][indentation_to_remove:]
        
        while True:
            if lines[-1]:
                break
            
            del lines[-1]
            if not lines:
                return None
        
        while True:
            if lines[0]:
                break
            
            del lines[0]
            continue
        
        
        documentation_parts = []
        index = 0
        line_count = len(lines)
        
        while True:
            documentation_parts.append('> ')
            documentation_parts.append(lines[index])
            
            index += 1
            if index == line_count:
                break
            
            documentation_parts.append('\n')
            continue
        
        return ''.join(documentation_parts)



from .configuration import *
from .constants import *
from .default import *
from .helpers import *
from .manager import *
from .scarletio_coroutine import *


__all__ = (
    *configuration.__all__,
    *constants.__all__,
    *default.__all__,
    *helpers.__all__,
    *manager.__all__,
    *scarletio_coroutine.__all__,
)



__all__ = ('FileRegistrationDoneEvent',)

from .base import EventBase
from .constants import IDENTIFIER_FILE_REGISTRATION_DONE


class FileRegistrationDoneEvent(EventBase):
    """
    Dispatched when all test files were registered.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_FILE_REGISTRATION_DONE



__all__ = (
    'apply_environments_for_file_at', 'set_global_environment', 'set_directory_environment', 'set_file_environment'
)

from os.path import basename as get_file_name, dirname as get_directory_name

from scarletio import get_last_module_frame

from .default import DefaultEnvironment


ENVIRONMENT_SCOPE_GLOBAL = 1
ENVIRONMENT_SCOPE_DIRECTORY = 2
ENVIRONMENT_SCOPE_FILE = 3


ENVIRONMENTS_BY_SCOPE = {}


def _set_environment_by_scope(environment, scope, detail):
    """
    Stores an environment.
    
    Parameters
    ----------
    environment : ``Environment``
        The environment to store.
    scope : `int`
        The scope of the storage.
    detail : `str`
        Additional detail for lookup.
        
    Raises
    ------
    TypeError
        - If `environment`'s type is incorrect.
    """
    _check_environment_type(environment)
    
    try:
        environment_by_detail = ENVIRONMENTS_BY_SCOPE[scope]
    except KeyError:
        environment_by_detail = {}
        ENVIRONMENTS_BY_SCOPE[scope] = environment_by_detail
    
    try:
        environments_by_identifier = environment_by_detail[detail]
    except KeyError:
        environments_by_identifier = {}
        environment_by_detail[detail] = environments_by_identifier
    
    environments_by_identifier[environment.identifier] = environment


def _get_last_module_file_path():
    """
    Gets the file name of the module from where the last function was called.
    
    Returns
    -------
    file_name : `None`, `str`
    """
    frame = get_last_module_frame()
    if (frame is None):
        return None
    
    spec = frame.f_globals.get('__spec__', None)
    if spec is None:
        return None
    
    return spec.origin


def set_global_environment(environment):
    """
    Sets an environment for global scope.
    
    Parameters
    ----------
    environment : ``Environment``
        The environment to set.
        
    Raises
    ------
    TypeError
        - If `environment`'s type is incorrect.
    """
    return _set_environment_by_scope(environment, ENVIRONMENT_SCOPE_GLOBAL, '')


def set_directory_environment(environment):
    """
    Sets an environment for directory scope.
    
    Parameters
    ----------
    environment : ``Environment``
        The environment to set.
    
    Raises
    ------
    RuntimeError
        - If not called from an `__init__.py` file.
    TypeError
        - If `environment`'s type is incorrect.
    """
    file_path = _get_last_module_file_path()
    if file_path is None:
        raise RuntimeError(
            f'Cannot set directory level environment, top level file could not be detected successfully.'
        )
    
    if get_file_name(file_path) != '__init__.py':
        raise RuntimeError(
            f'Cannot set directory level environment, top level file is not an `__init__.py` file.'
        )
    
    directory_name = get_directory_name(file_path)
    
    return _set_environment_by_scope(environment, ENVIRONMENT_SCOPE_DIRECTORY, directory_name)


def set_file_environment(environment):
    """
    Sets an environment for directory scope.
    
    Parameters
    ----------
    environment : ``Environment``
        The environment to set.
    
    Raises
    ------
    TypeError
        - If `environment`'s type is incorrect.
    """
    file_path = _get_last_module_file_path()
    if (file_path is None):
        raise RuntimeError(
            f'Cannot set file level environment, top level file could not be detected successfully.'
        )
    
    return _set_environment_by_scope(environment, ENVIRONMENT_SCOPE_FILE, file_path)


def _iter_environments_for_file_at(path):
    """
    Iterates  over the registered environments for the given `path`.
    
    This method is an iterable generator.
    
    Parameters
    ----------
    path : `str`
        The respective file's path.
    """
    # Global
    try:
        environment_by_detail = ENVIRONMENTS_BY_SCOPE[ENVIRONMENT_SCOPE_GLOBAL]
    except KeyError:
        pass
    else:
        try:
            environments_by_identifier = environment_by_detail['']
        except KeyError:
            pass
        else:
            yield from environments_by_identifier.values()
    
    # Directory
    try:
        environment_by_detail = ENVIRONMENTS_BY_SCOPE[ENVIRONMENT_SCOPE_DIRECTORY]
    except KeyError:
        pass
    else:
        directory_name = get_directory_name(path)
        
        try:
            environments_by_identifier = environment_by_detail[directory_name]
        except KeyError:
            pass
        else:
            yield from environments_by_identifier.values()
    
    # File
    try:
        environment_by_detail = ENVIRONMENTS_BY_SCOPE[ENVIRONMENT_SCOPE_FILE]
    except KeyError:
        pass
    else:
        try:
            environments_by_identifier = environment_by_detail[path]
        except KeyError:
            pass
        else:
            yield from environments_by_identifier.values()


def apply_environments_for_file_at(environment_manager, path):
    """
    Applies environments on the given manager for the given path.
    
    Parameters
    ----------
    environment_manager : ``EnvironmentManager``
        The environmental manager to alter.
    path : `str`
        The respective file's path.
    
    Returns
    -------
    environment_manager : ``EnvironmentManager``
    """
    return environment_manager.with_environment(*_iter_environments_for_file_at(path))


def _check_environment_type(environment):
    """
    Checks whether the given `environment` is really an environment.
    
    Parameters
    ----------
    environment : ``Environment``
    
    Raises
    ------
    TypeError
        - If `environment`'s type is incorrect.
    """
    if not isinstance(environment, DefaultEnvironment):
        # Check whether they forgot to instance it.
        # If we just drop a random type error, they wont find what they messed up.
        if isinstance(environment, type) and issubclass(environment, DefaultEnvironment):
            raise TypeError(
                f'You wanted to register an environment without instancing it, '
                f'got {environment.__name__}'
            )
        
        raise TypeError(
            f'`environment` can be `{DefaultEnvironment.__name__}`, '
            f'got {environment.__class__.__name__}; {environment!r}.'
        )



import sys

from scarletio import get_short_executable

from .main import execute_from_terminal


def __main__():
    """
    Executes vampytest from terminal.
    """
    parameters = sys.argv.copy()
    
    if parameters and ((parameters[0] == __file__) or (parameters[0] != get_short_executable())):
        del parameters[0]
    
    execute_from_terminal(parameters)


if __name__ == '__main__':
    __main__()



"""
Test assertions

Examples
--------
```
import vampytest


@vampytest.call_with('cake', 'cake')
def test_is_equal(p1, p2):
    vampytest.assert_equals(p1, p2)


@vampytest.returning('cake')
def test_returning_cake():
    return 'cake'
```
"""

from .aliases import *
from .wrapper_base import *
from .wrapper_call import *
from .wrapper_chainer import *
from .wrapper_conflict import *
from .wrapper_environment import *
from .wrapper_garbage_collect import *
from .wrapper_reverse import *
from .wrapper_skip import *
from .wrapper_skip_conditional import *


__all__ = (
    *aliases.__all__,
    *wrapper_base.__all__,
    *wrapper_call.__all__,
    *wrapper_chainer.__all__,
    *wrapper_conflict.__all__,
    *wrapper_environment.__all__,
    *wrapper_garbage_collect.__all__,
    *wrapper_reverse.__all__,
    *wrapper_skip.__all__,
    *wrapper_skip_conditional.__all__,
)



from .aliases import *
from .capture_output_context_manager import *
from .wrap_nothing import *


__all__ = (
    *aliases.__all__,
    *capture_output_context_manager.__all__,
    *wrap_nothing.__all__,
)



from .collection import *
from .file_system_entry import *
from .load_failure import *
from .test_file import *


__all__ = (
    *collection.__all__,
    *file_system_entry.__all__,
    *load_failure.__all__,
    *test_file.__all__,
)



__all__ = ('DefaultEnvironment',)

from scarletio import RichAttributeErrorBaseType

from ..handling import ResultState

from .constants import ENVIRONMENT_TYPE_DEFAULT


class DefaultEnvironment(RichAttributeErrorBaseType):
    """
    Implements a default environment to call tests in.
    
    Class Attributes
    ----------------
    identifier : `int` = `ENVIRONMENT_TYPE_DEFAULT`
        Represents for which environment the test is applicable for.
    """
    __slots__ = ()
    
    identifier = ENVIRONMENT_TYPE_DEFAULT
    
    def __new__(cls):
        """
        Creates a new environment.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the environment's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def run(self, test, positional_parameters, keyword_parameters):
        """
        Runs the defined test with the given parameters.
        
        Parameters
        ----------
        test : `FunctionType`
            The test to call
        positional_parameters : `list` of `object`
            Positional parameters to call the test with.
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters to call the test with.
        
        Returns
        -------
        result_state : ``ResultState``
            The result's product.
        """
        try:
            returned_value = test(*positional_parameters, **keyword_parameters)
        except BaseException as raised_exception:
            return ResultState().with_raise(raised_exception)
        
        return ResultState().with_return(returned_value)
    
    
    def shutdown(self):
        """
        Stops the environment.
        
        Called on environments when testing ended and the process should shut down.
        """
        pass



__all__ = ('TestingStartEvent',)

from .base import EventBase
from .constants import IDENTIFIER_TESTING_START


class TestingStartEvent(EventBase):
    """
    Dispatched when testing is started.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    """
    __slots__ = ()
    
    identifier = IDENTIFIER_TESTING_START



__all__ = ('mock_globals',)

from types import FunctionType


BUILTINS = {
    name: __builtins__.get(name, NotImplemented) for name in (
        'ArithmeticError', 'AssertionError', 'AttributeError', 'BaseException', 'BlockingIOError', 'BrokenPipeError',
        'BufferError', 'BytesWarning', 'ChildProcessError', 'ConnectionAbortedError', 'ConnectionError',
        'ConnectionRefusedError', 'ConnectionResetError', 'DeprecationWarning', 'EOFError', 'Ellipsis',
        'EnvironmentError', 'Exception', 'False', 'FileExistsError', 'FileNotFoundError', 'FloatingPointError',
        'FutureWarning', 'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 'IndentationError', 'IndexError',
        'InterruptedError', 'IsADirectoryError', 'KeyError', 'KeyboardInterrupt', 'LookupError', 'MemoryError',
        'ModuleNotFoundError', 'NameError', 'None', 'NotADirectoryError', 'NotImplemented', 'NotImplementedError',
        'OSError', 'OverflowError', 'PendingDeprecationWarning', 'PermissionError', 'ProcessLookupError',
        'RecursionError', 'ReferenceError', 'ResourceWarning', 'RuntimeError', 'RuntimeWarning', 'StopAsyncIteration',
        'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 'SystemExit', 'TabError', 'TimeoutError',
        'True', 'TypeError', 'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 'UnicodeError',
        'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning', 'ValueError', 'Warning', 'ZeroDivisionError',
        '__import__', 'abs', 'all', 'any', 'ascii', 'bin', 'bool', 'breakpoint', 'bytearray', 'bytes', 'callable',
        'chr', 'classmethod', 'compile', 'complex', 'copyright', 'credits', 'delattr', 'dict', 'dir', 'divmod',
        'enumerate', 'eval', 'exec', 'exit', 'filter', 'float', 'format', 'frozenset', 'getattr', 'globals', 'hasattr',
        'hash', 'help', 'hex', 'id', 'input', 'int', 'isinstance', 'issubclass', 'iter', 'len', 'license', 'list',
        'locals', 'map', 'max', 'memoryview', 'min', 'next', 'object', 'oct', 'open', 'ord', 'pow', 'print',
        'property', 'quit', 'range', 'repr', 'reversed', 'round', 'set', 'setattr', 'slice', 'sorted', 'staticmethod',
        'str', 'sum', 'super', 'tuple', 'type', 'vars', 'zip'
    )
}


def mock_globals(to_mock, recursion = -1, values = None, **keyword_parameters):
    """
    Mocks the globals of the given function.
    
    Parameters
    ----------
    to_mock : `FunctionType`
        The function to mock.
    recursion : `int` = `-1`, Optional
        Recursion level. Can be used to mock nested function if given as `> 1`.
    values : `None | dict<str, object>` = `None`, Optional
        Can be used to define the values to mock from a dictionary.
    **keyword_parameters : `dict<str, object>`
        Additional keyword values to define values as mock easier if their name is not an overlap with an actual
        parameter.
    
    Returns
    -------
    mocked : `instance<type<to_mock>>`
        The mocked function.
    """
    if values is not None:
        keyword_parameters.update(values)
    
    if isinstance(to_mock, FunctionType):
        return _mock_function_globals(to_mock, recursion, keyword_parameters)
    
    raise TypeError(
        f'Cannot mock {to_mock.__class__.__name__}; {to_mock!r}.'
    )


def _create_mocked_globals(potential_globals, recursion, old_globals, new_values):
    """
    Mocks the old globals dictionary returning the new ones.
    
    Parameters
    ----------
    potential_globals : `Generator`
        Generator over the potential globals we want to grab.
    recursion : `int`
        Recursion level.
    old_globals : `dict<str, object>`
        The old globals to mock.
    new_values : `dict<str, object>`
        The new values to mock with.
    
    Returns
    -------
    mew_globals : `dict<str, object>`
    """
    new_globals = {}
    recursion -= 1
    
    # There is a python error I cannot reproduce with simple tests, but apparently some shit tries to access
    # `__builtins__` over actual globals
    try:
        builtins = old_globals['__builtins__']
    except KeyError:
        pass
    else:
        new_globals['__builtins__'] = builtins
    
    for name in potential_globals:
        try:
            value = old_globals[name]
        except KeyError:
            try:
                value = BUILTINS[name]
            except KeyError:
                # Variable missing ?!
                continue
        
        try:
            value = new_values[name]
        except KeyError:
            if recursion > 0:
                if isinstance(value, FunctionType):
                    value = _mock_function_globals(value, recursion, new_values)
        
        new_globals[name] = value
    
    return new_globals


def _mock_function_globals(to_mock, recursion, new_values):
    """
    Mocks a function's globals.
    
    Parameters
    ----------
    to_mock : `FunctionType`
        The function to mock.
    recursion : `int`
        Recursion level.
    new_values : `dict<str, object>`
        The new values to mock with.
    
    Returns
    -------
    mocked : `FunctionType`
        The mocked function.
    """
    mocked = FunctionType(
        to_mock.__code__,
        _create_mocked_globals(_iter_potential_globals_of(to_mock), recursion, to_mock.__globals__, new_values),
        to_mock.__name__,
        to_mock.__defaults__,
        to_mock.__closure__,
    )
    mocked.__kwdefaults__ = to_mock.__kwdefaults__
    return mocked


def _iter_potential_globals_of(function):
    """
    Iterates over the potential globals of the given function.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    function : ``FunctionType`
        The function to check.
    
    Yields
    ------
    name : `str`
    """
    code_object = function.__code__
    
    local_variable_names = {*code_object.co_varnames}
    for name in code_object.co_names:
        if name not in local_variable_names:
            yield name



__all__ = ('AssertionBase',)

from scarletio import RichAttributeErrorBaseType

from .assertion_states import ASSERTION_STATE_NONE, get_assertion_state_name


class AssertionBase(RichAttributeErrorBaseType):
    """
    Base class for conditions.
    
    Attributes
    ----------
    state : `str`
        The condition's state.
    """
    __slots__ = ('state',)
    
    
    def __new__(cls):
        """
        Creates a new condition.
        """
        self = object.__new__(cls)
        self.state = ASSERTION_STATE_NONE
        return self
    
    
    def _build_repr_parts_into(self, into):
        """
        Representation builder helper.
        
        Parameters
        ----------
        into : `list` of `str`
            List of strings to build the representation into.
        
        Returns
        -------
        into : `list` of `str`
        """
        return into
    
    
    def __repr__(self):
        """Returns the condition's representation."""
        repr_parts = ['<', self.__class__.__name__]
        repr_parts.append(' state = ')
        repr_parts.append(get_assertion_state_name(self.state))
        repr_parts = self._build_repr_parts_into(repr_parts)
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def render_failure_message_parts_into(self, failure_message_parts):
        """
        Renders the assertion into the given list.
        
        Parameters
        ----------
        failure_message_parts : `list` of `str`
            A list to extend with the rendered strings.
        
        Returns
        -------
        failure_message_parts : `list` of `str`
        """
        return failure_message_parts



import pathlib, re
from ast import literal_eval
from setuptools import setup

HERE = pathlib.Path(__file__).parent

# Lookup version
version_search_pattern = re.compile('^__version__[ ]*=[ ]*((?:\'[^\']+\')|(?:\"[^\"]+\"))[ ]*$', re.M)
parsed = version_search_pattern.search((HERE / 'vampytest' / '__init__.py').read_text())
if parsed is None:
    raise RuntimeError('No version found in `__init__.py`.')

version = literal_eval(parsed.group(1))

# Lookup readme
README = (HERE / 'README.md').read_text('utf-8')

setup(
    author = 'HuyaneMatsu',
    author_email = 're.ism.tm@gmail.com',
    classifiers = [
        'Development Status :: 2 - Pre-Alpha',
        
        'Intended Audience :: Developers',
        
        'Operating System :: OS Independent',
        
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        
        'Topic :: Software Development :: Testing',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    description = 'A vampy test framework',
    entry_points = {
        'console_scripts': [
            'vampytest = vampytest.__main__:__main__'
        ],
    },
    include_package_data = True,
    install_requires = [
        'scarletio>=1.0.64',
    ],
    license = 'DBAD',
    long_description = README,
    long_description_content_type = 'text/markdown',
    name = 'vampytest',
    packages = [
        'vampytest',
        'vampytest.core',
        'vampytest.core.assertions',
        'vampytest.core.contexts',
        'vampytest.core.environment',
        'vampytest.core.event_handling',
        'vampytest.core.event_handling.rendering_helpers',
        'vampytest.core.events',
        'vampytest.core.file',
        'vampytest.core.handling',
        'vampytest.core.helpers',
        'vampytest.core.mocking',
        'vampytest.core.result',
        'vampytest.core.result.reports',
        'vampytest.core.runner',
        'vampytest.core.utils',
        'vampytest.core.wrappers',
        'vampytest.main',
    ],
    python_requires = '>=3.6',
    url = 'https://github.com/HuyaneMatsu/vampytest',
    version = version,
)



__all__ = ()

import reprlib

from scarletio import RichAttributeErrorBaseType, render_exception_into

from .test_file import __file__ as VAMPYTEST_TEST_FILE_PATH


def _ignore_module_import_frame(frame):
    """
    Ignores the frame, where the test file was imported.
    
    Parameters
    ----------
    frame : ``frameProxyBase``
        The frame to check.
    
    Returns
    -------
    should_show_frame : `bool`
        Whether the frame should be shown.
    """
    should_show_frame = True
    
    file_name = frame.file_name
    name = frame.name
    line = frame.line
    
    if file_name == VAMPYTEST_TEST_FILE_PATH:
        if name == '_try_load_module':
            if line == '__import__(import_route)':
                should_show_frame = False
    
    return should_show_frame


class TestFileLoadFailure(RichAttributeErrorBaseType):
    """
    Represents an occurred failure meanwhile loading a test file.
    
    Attributes
    ----------
    exception_message : `str`
        Exception message with traceback.
    path : `str`
        The files path which failed to load.
    """
    __slots__ = ('path', 'exception_message')
    
    def __new__(cls, test_file, exception):
        """
        Parameters
        ----------
        test_file : ``TestFile``
            The test file which failed to load.
        exception : `BaseException`
            The occurred exception.
        """
        exception_message_parts = []
        render_exception_into(exception, exception_message_parts, filter=_ignore_module_import_frame)
        exception_message = ''.join(exception_message_parts)
        
        self = object.__new__(cls)
        self.exception_message = exception_message
        self.path = test_file.path
        return self
    
    
    def __repr__(self):
        """Returns the test loading failure's representation."""
        return f'<{self.__class__.__name__} of {reprlib.repr(self.path)}>'



__all__ = (
    'call_from', 'call_with', 'in_environment', 'raising', 'returning', 'reverse', 'revert', 'skip', 'skip_if',
    'with_gc',
)

import warnings

from .wrapper_call import WrapperCalling
from .wrapper_call_from import WrapperCallingFrom
from .wrapper_environment import WrapperEnvironment
from .wrapper_garbage_collect import WrapperGarbageCollect
from .wrapper_reverse import WrapperReverse
from .wrapper_skip import WrapperSkip
from .wrapper_skip_conditional import WrapperSkipConditional

call_from = WrapperCallingFrom.calling_from_constructor
call_with = WrapperCalling.call_with_constructor
in_environment = WrapperEnvironment
raising = WrapperCalling.raising_constructor
returning = WrapperCalling.returning_constructor
reverse = WrapperReverse
skip = WrapperSkip
skip_if = WrapperSkipConditional
with_gc = WrapperGarbageCollect


def revert(*pp, **kp):
    """
    Deprecated and will be removed in 2024 February. Please use ``reverse`` instead.
    """
    warnings.warn(
        (
            f'`revert` is deprecated and will be removed in 2024 February. '
            f'Please use `.reverse` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )
    return reverse(*pp, **kp)



__all__ = ('EventBase',)

from scarletio import RichAttributeErrorBaseType, copy_docs

from .constants import IDENTIFIER_NONE


class EventBase(RichAttributeErrorBaseType):
    """
    Represents a dispatched event by a test runner.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test running context.
    """
    __slots__ = ('context',)
    
    identifier = IDENTIFIER_NONE
    
    def __new__(cls, context):
        """
        Creates a new event instance.
        
        Parameters
        ----------
        context : ``RunnerContext``
            The respective test running context.
        """
        self = object.__new__(cls)
        self.context = context
        return self
    
    
    def __repr__(self):
        """Returns the representation of the event."""
        return f'<{self.__class__.__name__}>'


class FileEventBase(EventBase):
    """
    Represents a dispatched event by a test runner.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test running context.
    file : ``TestFile``
        The respective test file.
    """
    __slots__ = ('file',)
    
    def __new__(cls, context, file):
        """
        Creates a new file event instance.
        
        Parameters
        ----------
        context : ``RunnerContext``
            The respective test running context.
        file : ``TestFile``
            The respective test file.
        """
        self = EventBase.__new__(cls, context)
        self.file = file
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} file = {self.file!r}>'


class ResultEventBase(EventBase):
    """
    Represents a dispatched event by a test runner.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test running context.
    result : ``Result``
        The respective result.
    """
    __slots__ = ('result',)
    
    def __new__(cls, context, result):
        """
        Creates a new file event instance.
        
        Parameters
        ----------
        context : ``RunnerContext``
            The respective test running context.
        result : ``Result``
            The respective result.
        """
        self = EventBase.__new__(cls, context)
        self.result = result
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        return f'<{self.__class__.__name__} result = {self.result!r}>'



__all__ = ('EnvironmentManager', )

from scarletio import RichAttributeErrorBaseType

from .helpers import get_function_environment_identifier
from .default import DefaultEnvironment
from .scarletio_coroutine import ScarletioCoroutineEnvironment


DEFAULT_ENVIRONMENT_TYPES = (
    DefaultEnvironment,
    ScarletioCoroutineEnvironment,
)


class EnvironmentManager(RichAttributeErrorBaseType):
    """
    Represents an environment manager managing testing environments.
    
    Attributes
    ----------
    _environments_by_identifier : `dict` of (`int`, ``DefaultEnvironment``) items
    """
    __slots__ = ('_environments_by_identifier',)
    
    def __new__(cls):
        """
        Creates a new environment manager.
        """
        self = object.__new__(cls)
        self._environments_by_identifier = {}
        return self
    
    def __repr__(self):
        """Returns the testing environment's representation."""
        return f'<{self.__class__.__name__} environments_by_identifier = {self._environments_by_identifier!r}>'
    
    
    def get_environment_for_test(self, test):
        """
        Gets environment for the given test.
        
        Parameters
        ----------
        test : `FunctionType`
            Test function.
        
        Returns
        -------
        environment : ``DefaultEnvironment``
        
        Raises
        ------
        NotImplementedError
        """
        identifier = get_function_environment_identifier(test)
        try:
            environment = self._environments_by_identifier[identifier]
        except KeyError:
            raise NotImplementedError(
                f'No environment defined for type: {identifier}'
            ) from None
        
        return environment
    
    
    def with_environment(self, *environments):
        """
        Copies the environment manager with the given environment.
        
        Parameters
        ----------
        *environments : ``DefaultEnvironment``
            The environment to copy self with-
        
        Returns
        -------
        new : ``EnvironmentManager``
        """
        if not environments:
            return self
        
        new = self.copy()
        
        environments_by_identifier = new._environments_by_identifier
        for environment in environments:
            environments_by_identifier[environment.identifier] = environment
        
        return new
    
    
    def copy(self):
        """
        Copies the environment manager.
        
        Returns
        -------
        new : ``EnvironmentManager``
        """
        environments_by_identifier = self._environments_by_identifier.copy()
        
        new = object.__new__(type(self))
        new._environments_by_identifier = environments_by_identifier
        return new


    def populate(self):
        """
        Tries to auto populate the environment manager.
        
        Returns
        -------
        self : ``EnvironmentManager``
        """
        environments_by_identifier = self._environments_by_identifier
        
        for environment_type in DEFAULT_ENVIRONMENT_TYPES:
            identifier = environment_type.identifier
            if identifier not in environments_by_identifier:
                environments_by_identifier[identifier] = environment_type()
        
        return self



__version__ = '0.0.15'

try:
    import scarletio
except ImportError as err:
    raise ImportError(
        'Couldn\'t import scarletio. \n'
        'Are you sure it\'s installed and available on your PYTHONPATH environment variable?\n'
        'Did you forget to activate a virtual environment?\n'
    ) from err

del scarletio


from .core import *

__all__ = core.__all__



__all__ = (
    'assert_', 'assert_contains', 'assert_eq', 'assert_equals', 'assert_false', 'assert_id', 'assert_identical',
    'assert_in', 'assert_instance', 'assert_is', 'assert_is_not', 'assert_ne', 'assert_not', 'assert_not_contains',
    'assert_not_eq', 'assert_not_equals', 'assert_not_id', 'assert_not_identical', 'assert_not_in', 'assert_not_is',
    'assert_raises', 'assert_subtype', 'assert_true'
)


from .assertion_contains import AssertionContains
from .assertion_equals import AssertionEquals
from .assertion_false import AssertionValueEvaluationFalse
from .assertion_identical import AssertionIdentical
from .assertion_instance import AssertionInstance
from .assertion_not_contains import AssertionNotContains
from .assertion_not_equals import AssertionNotEquals
from .assertion_not_identical import AssertionNotIdentical
from .assertion_raising import AssertionRaising
from .assertion_subtype import AssertionSubtype
from .assertion_true import AssertionValueEvaluationTrue


assert_ = AssertionValueEvaluationTrue
assert_contains = AssertionContains
assert_eq = AssertionEquals
assert_equals = AssertionEquals
assert_false = AssertionValueEvaluationFalse
assert_id = AssertionIdentical
assert_identical = AssertionIdentical
assert_in = AssertionContains
assert_instance = AssertionInstance
assert_is = AssertionIdentical
assert_is_not = AssertionNotIdentical
assert_ne = AssertionNotEquals
assert_not = AssertionValueEvaluationFalse
assert_not_contains = AssertionNotContains
assert_not_eq = AssertionNotEquals
assert_not_equals = AssertionNotEquals
assert_not_id = AssertionNotIdentical
assert_not_identical = AssertionNotIdentical
assert_not_in = AssertionNotContains
assert_not_is = AssertionNotIdentical
assert_raises = AssertionRaising
assert_subtype = AssertionSubtype
assert_true = AssertionValueEvaluationTrue



__all__ = ('Result',)

from scarletio import RichAttributeErrorBaseType, export

from .reports import (
    ReportFailureAsserting, ReportFailureParameterMismatch, ReportFailureRaising, ReportFailureReturning, ReportOutput
)


@export
class Result(RichAttributeErrorBaseType):
    """
    Represents a test's result.
    
    Attributes
    ----------
    case : ``TestCase``
        The parent test case creating this test group.
    conflict : `None`, ``WrapperConflict``
        Conflicts between test wrappers.
    continuous : `bool`
        Whether this is is not the last test of the case.
    handle : ``Handle``
        The test's handle running the test.
    reports : `None`, `list` of ``ReportBase``
        Test reports.
    reversed : `bool`
        Whether the test result is reversed.
    skipped : `bool`
        Whether the test is skipped.
    
    Utility Methods
    - ``.is_skipped``
    - ``.is_passed``
    - ``.is_failed``
    - ``.is_conflicted``
    - ``.is_informal``
    - ``.iter_report_messages``
    """
    __slots__ = ('case', 'conflict', 'continuous', 'handle', 'reports', 'reversed', 'skipped')
    
    def __new__(cls, case):
        """
        Creates a new test result.
        
        Parameters
        ----------
        case : ``TestCase``
            The parent test case creating this test group.
        """
        self = object.__new__(cls)
        self.case = case
        self.conflict = None
        self.continuous = False
        self.handle = None
        self.reports = None
        self.reversed = case.do_reverse()
        self.skipped = False
        return self
    
    
    def __repr__(self):
        """Returns the test result's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        
        conflict = self.conflict
        if (conflict is None):
            field_added = False
        
        else:
            repr_parts.append(' conflict = ')
            repr_parts.append(repr(repr_parts))
            field_added = True
        
        reports = self.reports
        if (reports is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' reports = ')
            repr_parts.append(repr(reports))
        
        if self.reversed:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' reversed')
        
        if self.skipped:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' skipped')
        
        if self.continuous:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' continuous')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def with_handle(self, handle):
        """
        Sets handle.
        
        Parameters
        ----------
        handle : ``Handle``
            The test's handle running the test.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        self.handle = handle
        return self
    
    
    def with_conflict(self, conflict):
        """
        Adds the conflict to the test result.
        
        Parameters
        ----------
        conflict : ``WrapperConflict``
            Conflicts between test wrappers.
            
        Returns
        -------
        self : `instance<type<self>>`
        """
        self.conflict = conflict
        return self
    
    
    def as_skipped(self):
        """
        Marks the test result as skipped.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        self.skipped = True
        return self
    
    
    def as_continuous(self):
        """
        Marks the test result as continuous.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        self.continuous = True
        return self
    
    
    def with_exception(self, expected_exceptions, received_exception, exact_type):
        """
        Adds unexpected exception as test result.
        
        Parameters
        ----------
        expected_exceptions : `None`, `set` of `BaseException`
            A set of the expected exceptions.
        received_exception : `None`, ``BaseException``
            The received exception.
        exact_type : `bool`
            Whether exception subclasses are disallowed.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        report = ReportFailureRaising(expected_exceptions, received_exception, exact_type)
        self._add_report(report)
        return self
    
    
    def with_return(self, expected_value, received_value):
        """
        Adds return value mismatch as test result.
        
        Parameters
        ----------
        expected_value : `object`
            The expected return value.
        received_value : `object`
            The received return value.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        report = ReportFailureReturning(expected_value, received_value)
        self._add_report(report)
        return self
    
    
    def with_assertion(self, assertion_exception):
        """
        Adds assertion failing as test result.
        
        Parameters
        ----------
        assertion_exception : ``AssertionException``
            The failed assertion.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        report = ReportFailureAsserting(assertion_exception)
        self._add_report(report)
        return self
    
    
    def with_parameter_mismatch(self, parameter_mismatch):
        """
        Adds a parameter mismatch as a test result.
        
        Parameters
        ----------
        parameter_mismatch : ``ParameterMismatch``
            The parameter mismatch to add.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        report = ReportFailureParameterMismatch(parameter_mismatch)
        self._add_report(report)
        return self
    
    
    def with_output(self, output):
        """
        Adds captured output as test result. 
        
        Parameters
        ----------
        output : `str`
            Output to add.
        
        Returns
        -------
        self : `instance<type<self>>`
        """
        report = ReportOutput(output)
        self._add_report(report)
        return self
    
    
    def _add_report(self, report):
        """
        Ads a report to the test result.
        
        Parameters
        ----------
        report : ``ReportBase``
            The test report to add.
        """
        reports = self.reports
        if (reports is None):
            reports = []
            self.reports = reports
        
        reports.append(report)
    
    
    def is_skipped(self):
        """
        Returns whether the represented test case was skipped.
        
        Returns
        -------
        is_skipped : `bool`
        """
        return self.skipped
    
    
    def is_passed(self):
        """
        Returns whether the test result passed.
        
        Returns
        -------
        is_failed : `bool`
        """
        if self.skipped:
            return True
        
        if (self.conflict is not None):
            return False
        
        return (self.get_failure_report() is None) != self.reversed
    
    
    def is_failed(self):
        """
        Returns whether the test result failed.
        
        Returns
        -------
        is_failed : `bool`
        """
        if self.skipped:
            return False
        
        if (self.conflict is not None):
            return True
        
        return (self.get_failure_report() is None) == self.reversed
    
    
    def is_conflicted(self):
        """
        Returns whether the respective test case conflicted.
        
        Returns
        -------
        is_conflicted : `bool`
        """
        return (self.conflict is not None)
    
    
    def is_informal(self):
        """
        Returns whether the test result holds only informal results.
        
        Returns
        -------
        is_informal : `bool`
        """
        if self.skipped:
            return False
        
        if (self.conflict is not None):
            return False
        
        has_informal = False
        
        for report in self.iter_reports():
            if report.is_failure():
                return False
            
            if report.is_informal():
                has_informal = True
        
        return has_informal
        
    
    def is_last(self):
        """
        Returns whether the result is the last of the test case. Can be used when rendering test tree.
        
        Returns
        -------
        is_last : `bool`
        """
        return (not self.continuous)
    
    
    def get_modifier_parameters(self):
        """
        Returns the test's modifier parameters.
        
        Returns
        -------
        modifier_parameters : `None | (None | list<object>, None | dict<str, object)`
        """
        handle = self.handle
        if handle is None:
            return None
        
        final_call_state = handle.final_call_state
        if final_call_state is None:
            return None
        
        if not final_call_state:
            return None
        
        return final_call_state.positional_parameters, final_call_state.keyword_parameters
    
    
    def iter_reports(self):
        """
        Iterates over the reports of the test result.
        
        Yields
        ------
        report : ``ReportBase``
        """
        reports = self.reports
        if reports is not None:
            yield from reports
    
    
    def get_failure_report(self):
        """
        Gets the first failure report.
        
        Returns
        -------
        report : `None`, ``ReportBase``
        """
        for report in self.iter_reports():
            if report.is_failure():
                return report
    
    
    def get_output_report(self):
        """
        Gets the first output report.
        
        Returns
        -------
        report : `None`, ``ReportOutput``
        """
        for report in self.iter_reports():
            if isinstance(report, ReportOutput):
                return report



__all__ = ()

from scarletio import RichAttributeErrorBaseType


class ContextBase(RichAttributeErrorBaseType):
    """
    Base test context.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new test context.
        """
        return object.__new__(cls)
    
    
    def __repr__(self):
        """Returns the context's representation."""
        return ''.join(['<', self.__class__.__name__, '>'])
    
    
    def start(self):
        """
        Starts the test context. If the test should already end returns a ``Result`` instance.
        
        Returns
        -------
        result : `None`, ``Result``
        """
        return None
    
    
    def enter(self, call_state):
        """
        Enters the context.
        
        Parameters
        ----------
        call_state : ``CallState``
            Call state of the test to modify.
            
        Returns
        -------
        result : `None`, ``Result``
            Test result to propagate.
        call_state : `None`, ``CallState``
            New call state to use.
        """
        return (None, None)
    
    
    def exit(self, result_state):
        """
        Exists the context.
        
        Parameters
        ----------
        result_state : ``ResultState``
            Result state of the test to modify.
        
        Returns
        -------
        result : `None`, ``Result``
            Test result to propagate.
        result_state : ``ResultState``
            New result state to use.
        """
        return (None, None)
    
    
    def close(self, result):
        """
        Closes the test context.
        
        Parameters
        ----------
        result : `None`, ``Result``
            Result of the test. Passed as `None` if an exception occurred while handling the test.
        """
        return None



"""
Test wrappers used to enhance our test experience.

Examples
--------
```
import vampytest

def test_equals():
    vampytest.assert_eq(5, 5)


def test_raises():
    with vampytest.raises(TypeError):
        'nice' + 69

def test_contains():
    vampytest.assert_in(range(7), 5)
```
"""
from .aliases import *
from .assertion_base import *
from .assertion_conditional_base import *
from .assertion_contains import *
from .assertion_equals import *
from .assertion_false import *
from .assertion_identical import *
from .assertion_instance import *
from .assertion_not_contains import *
from .assertion_not_equals import *
from .assertion_not_identical import *
from .assertion_raising import *
from .assertion_states import *
from .assertion_subtype import *
from .assertion_true import *
from .exceptions import *


__all__ = (
    *aliases.__all__,
    *assertion_raising.__all__,
    *assertion_base.__all__,
    *assertion_conditional_base.__all__,
    *assertion_contains.__all__,
    *assertion_equals.__all__,
    *assertion_false.__all__,
    *assertion_identical.__all__,
    *assertion_instance.__all__,
    *assertion_not_contains.__all__,
    *assertion_not_equals.__all__,
    *assertion_not_identical.__all__,
    *assertion_states.__all__,
    *assertion_subtype.__all__,
    *assertion_true.__all__,
    *exceptions.__all__,
)



__all__ = ()


def try_match_exception(expected_exceptions, received_exception, accept_subtypes, where):
    """
    Checks whether the received exception matches the preset ones.
    
    Parameters
    ----------
    expected_exceptions : `set` of `BaseException`
        The expected exceptions.
    received_exception : `BaseException`
        The received exception.
    accept_subtypes : `bool`
        Whether sub classes are allowed.
    where : `None`, `callable`
        Additional check to check the raised exception.
    
    Returns
    -------
    exception_matched : `bool`
    """
    for expected_exception in expected_exceptions:
        if isinstance(expected_exception, type):
            exception_type = expected_exception
            exception_value = None
        
        else:
            exception_type = type(expected_exception)
            exception_value = expected_exception
        
        
        if accept_subtypes:
            if not isinstance(received_exception, exception_type):
                continue
        
        else:
            if type(received_exception) is not exception_type:
                continue
        
        if (exception_value is not None) and (exception_value.args != received_exception.args):
            continue
        
        if (where is not None) and (not where(received_exception)):
            continue
        
        exception_matched = True
        break
    
    else:
        exception_matched = False
    
    return exception_matched



from .base import *
from .calling import *
from .garbage_collect import *
from .output_capturing import *


__all__ = (
    *base.__all__,
    *calling.__all__,
    *garbage_collect.__all__,
    *output_capturing.__all__,
)



__all__ = ('_',)


def _(value):
    """
    Wraps nothing and returns the value.
    
    Since python does not allow using arbitrary expression in decorators, it can be used to wrap the expression.
    
    Parameters
    ----------
    value : `object`
        The value to return.
    
    Returns
    -------
    value : `object`
    """
    return value



__all__ = ('SourceLoadFailureEvent',)

from scarletio import copy_docs

from .base import EventBase
from .constants import IDENTIFIER_SOURCE_LOAD_FAILURE


class SourceLoadFailureEvent(EventBase):
    """
    Dispatched when a test file's all tests ran.
    
    Attributes
    ----------
    context : ``RunnerContext``
        The respective test runner context.
    exception : ``BaseException``
        The occurred exception.
    source : `str`
        The source's name.
    """
    __slots__ = ('exception', 'source')
    
    identifier = IDENTIFIER_SOURCE_LOAD_FAILURE

    def __new__(cls, context, source, exception):
        """
        Creates a new file sy instance.
        
        Parameters
        ----------
        context : ``RunnerContext``
            The respective test running context.
        source : `str`
            The source's name.
        exception : ``BaseException``
            The occurred exception.
        """
        self = object.__new__(cls)
        self.context = context
        self.exception = exception
        self.source = source
        return self
    

    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' source = ')
        repr_parts.append(repr(self.source))
        
        repr_parts.append(', exception = ')
        repr_parts.append(repr(self.exception))
        
        repr_parts.append('>')
        return ''.join(repr_parts)



__all__ = ()

from scarletio import create_ansi_format_code


COLOR_FAIL = create_ansi_format_code(foreground_color = (255, 0, 0))
COLOR_PASS = create_ansi_format_code(foreground_color = (0, 255, 0))
COLOR_SKIP = create_ansi_format_code(foreground_color = (0, 255, 255))
COLOR_UNKNOWN = create_ansi_format_code(foreground_color = (255, 0, 255))
COLOR_RESET = create_ansi_format_code()

COLOR_PATH = create_ansi_format_code(foreground_color = (226, 153, 255))



__all__ = ()


def un_nest(value):
    """
    Un-nests the given value into a tuple.
    
    Parameters
    ----------
    value : `tuple<object>`, `object`
        The value to un-nest.
    
    Returns
    -------
    value : `tuple<object>`
    """
    return {*iter_un_nest(value)}


def iter_un_nest(value):
    """
    Un-nests the given tuple.
    
    This function is an iterable generator.
    
    Parameters
    ----------
    value : `tuple<object>`, `object`
        The value to un-nest.
    
    Yields
    ------
    value : `object`
    """
    to_do = [value]
    
    while to_do:
        value = to_do.pop()
        if isinstance(value, tuple):
            to_do.extend(value)
            continue
        
        yield value
        continue


def un_nest_exceptions(exceptions):
    """
    Un-nests the given `exceptions`.
    
    Parameters
    ----------
    exceptions : `BaseException | type<BaseException> | tuple<BaseException, <type<BaseException>, ...>`
        The exceptions to un-nest.
    
    Returns
    -------
    exceptions : `set` of (`BaseException`, `type<BaseException>`)
    
    Raises
    ------
    TypeError
        If an element's type is incorrect.
    """
    exceptions = un_nest(exceptions)
    
    for exception in exceptions:
        if isinstance(exception, BaseException):
            continue
        
        if isinstance(exception, type) and issubclass(exception, BaseException):
            continue
        
        raise TypeError(
            f'`exception` can be `BaseException` or its instance, `tuple` of (`BaseException`, ...)`, '
            f'got {exception.__class__.__name__}; {exception!r}.'
        )
    
    return exceptions


def un_nest_types(types):
    """
    Un-nests the given `types`.
    
    Parameters
    ----------
    types : `type<type> | tuple<type<type>, ...>`
        The types to un-nest.
    
    Returns
    -------
    types : `set` of `type`
    
    Raises
    ------
    TypeError
        If an element's type is incorrect.
    """
    types = un_nest(types)
    
    for type_value in types:
        if not isinstance(type_value, type):
            raise TypeError(
                f'`type` can be `type`, `tuple` of (`type`, ...)`, '
                f'got {type_value.__class__.__name__}; {type_value!r}.'
            )
    
    return types



__all__ = ('ReportFailureAsserting',)

from scarletio import copy_docs

from .base import ReportBase


class ReportFailureAsserting(ReportBase):
    """
    Test failure representing a failed assertion.
    
    Attributes
    ----------
    handle : ``Handle``
        The test's handle running the test.
    """
    __slots__ = ('assertion_exception',)
    
    def __new__(cls, assertion_exception):
        """
        Creates a new assertion test failure.
        
        Parameters
        ----------
        assertion_exception : ``AssertionException``
            The failed assertion.
        """
        self = object.__new__(cls)
        self.assertion_exception = assertion_exception
        return self
    
    
    @copy_docs(ReportBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' assertion = ')
        repr_parts.append(repr(self.assertion_exception))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ReportBase.is_failure)
    def is_failure(self):
        return True
